#!/usr/bin/env python3
#
# ############################################################################
# Copyright (c) 2025-2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import sys
import os
import argparse
from collections import defaultdict
from traceback import print_exc
import pandas
import logging
import gzip
import gffutils

try:
    from .error_codes import IsoQuantExitCode
    from .file_naming import mtx_matrix_file, mtx_features_file, mtx_barcodes_file
except ImportError:
    sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
    from error_codes import IsoQuantExitCode
    from file_naming import mtx_matrix_file, mtx_features_file, mtx_barcodes_file


GROUP_COUNT_CUTOFF = 100


logger = logging.getLogger('IsoQuant')


def set_logger(logger_instance):
    logger_instance.setLevel(logging.INFO)
    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.INFO)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger_instance.addHandler(ch)


def convert_to_matrix(input_linear_counts, output_file_path, feature_id_to_name_dict=None,
                      gzipped=False, feature_type='gene', convert_to_tpm=False, usable_reads_per_group=None,
                      max_groups: int = 0) -> int:
    """Convert linear counts to full matrix format.

    Returns:
        Number of groups found in the input file, or 0 on error.
    """
    logger.info("Converting %s to full matrix" % input_linear_counts)
    cols = len(open(input_linear_counts).readline().strip().split('\t'))
    if cols != 3:
        logger.error("Unexpected number of columns in %s: %d" % (input_linear_counts, cols))
        return 0

    # Quick group count check before expensive read+pivot
    if max_groups > 0:
        groups = set()
        with open(input_linear_counts) as f:
            next(f)  # skip header
            for line in f:
                parts = line.split('\t')
                if len(parts) >= 2:
                    groups.add(parts[1])
        num_groups = len(groups)
        if num_groups > max_groups:
            logger.info("Skipping full matrix conversion: %d groups exceeds limit of %d. "
                         "Use '--counts_format matrix' to force conversion." % (num_groups, max_groups))
            return num_groups

    df = pandas.read_csv(input_linear_counts, delimiter='\t', header=None, skiprows=1, keep_default_na=False,
                         names=['gene_id', 'group_id', 'count'],
                         dtype={'gene_id': str, 'group_id': str, 'count': float})
    count_matrix = df.pivot(index='gene_id', columns='group_id', values='count')

    normalization_factors = {g: 1.0 for g in df['group_id'].unique()} if not convert_to_tpm \
        else get_normalization_factors(df, usable_reads_per_group)

    num_groups = len(count_matrix.columns)

    output_file_path += ".tsv"
    output_file_path += ".gz" if gzipped else ""
    with gzip.open(output_file_path, 'wt') if gzipped else open(output_file_path, 'w') as outfile:
        # Write the header with group_ids
        columns = list(sorted(count_matrix.columns))
        if num_groups > GROUP_COUNT_CUTOFF:
            logger.warning("You have %d groups in your matrix, conversion might take a lot of time and the output file can be very large" % num_groups)
        outfile.write(feature_type + '_id\t' + '\t'.join(columns) + '\n')

        # Iterate over the DataFrame and write rows to the file
        for gene_id, row in count_matrix.iterrows():
            # Create a list of values, replacing NaNs with zeroes
            values = [str(row[col] * normalization_factors[col]) if pandas.notna(row[col]) else '0' for col in columns]
            gene_name = gene_id if feature_id_to_name_dict is None else feature_id_to_name_dict.get(gene_id, gene_id)
            outfile.write(gene_name + '\t' + '\t'.join(values) + '\n')
    logger.info("Matrix was saved to %s" % output_file_path)
    return num_groups


def convert_to_mtx(input_linear_counts, output_file_prefix, feature_id_to_name=None,
                   gzipped=False, convert_to_tpm=False, usable_reads_per_group=None):
    logger.info("Converting %s to MTX format" % input_linear_counts)
    cols = len(open(input_linear_counts).readline().strip().split('\t'))
    if cols != 3:
        logger.error("Unexpected number of columns in %s: %d" % (input_linear_counts, cols))
        return
    
    df = pandas.read_csv(input_linear_counts, delimiter='\t', header=None, skiprows=1, keep_default_na=False,
                         names=['gene_id', 'group_id', 'count'],
                         dtype={'gene_id': str, 'group_id': str, 'count': float})
    unique_genes = list(sorted(df['gene_id'].unique()))
    unique_groups = list(sorted(df['group_id'].unique()))
    gene_index_map = {gene_id: idx + 1 for idx, gene_id in enumerate(unique_genes)}
    group_index_map = {group_id: idx + 1 for idx, group_id in enumerate(unique_groups)}

    # Write the count matrix to an MTX file
    mtx_file = mtx_matrix_file(output_file_prefix)
    features_file = mtx_features_file(output_file_prefix)
    barcodes_files = mtx_barcodes_file(output_file_prefix)

    normalization_factors = {g: 1.0 for g in df['group_id'].unique()} if not convert_to_tpm \
        else get_normalization_factors(df, usable_reads_per_group)

    with open(barcodes_files, 'w') as bc_out:
        for group in unique_groups:
            bc_out.write(f"{group}\n")

    with open(features_file, 'w') as ft_out:
        for gene_id in unique_genes:
            gene_name = feature_id_to_name.get(gene_id, gene_id) if feature_id_to_name is not None else gene_id
            ft_out.write(f"{gene_id}\t{gene_name}\n")

    with gzip.open(mtx_file + ".gz", 'wt') if gzipped else open(mtx_file, 'w') as mtx_out:
        # Write the header
        mtx_out.write("%%MatrixMarket matrix coordinate real general\n")
        mtx_out.write(f"{len(unique_genes)} {len(unique_groups)} {df.shape[0]}\n")

        # Write the data in coordinate format
        for _, row in df.iterrows():
            gene_id = row['gene_id']
            group_id = row['group_id']
            count = row['count'] * normalization_factors[group_id]
            row_index = gene_index_map[gene_id]
            col_index = group_index_map[group_id]
            mtx_out.write(f"{row_index} {col_index} {count}\n")
    logger.info("Matrix was saved to 3 files with prefix %s" % output_file_prefix)


def get_normalization_factors(counts, usable_reads_per_group=None):
    if not usable_reads_per_group:
        total_group_counts = defaultdict(float)
        for _, row in counts.iterrows():
            group_id = row['group_id']
            count = row['count']
            total_group_counts[group_id] += count
    else:
        total_group_counts = usable_reads_per_group

    return {group_id: 1000000.0 / total_group_counts[group_id] for group_id in total_group_counts.keys()}


def load_gene_name_map(genedb):
    logger.info("Loading annotation from %s" % genedb)
    db = gffutils.FeatureDB(genedb, keep_order=True)
    transcript2gene = {}
    transcript_names = {}
    gene_names = {}

    for g in db.all_features(featuretype='gene'):
        gene_name = g.id
        if "gene_name" in g.attributes:
            gene_name = g["gene_name"][0]
        gene_names[g.id] = gene_name

    for t in db.all_features(featuretype=('transcript', 'mRNA')):
        if "gene_id" in t.attributes:
            gene_id = t["gene_id"][0]
        elif "Parent" in t.attributes:
            gene_id = t["Parent"][0]
        else:
            gene_id = "unknown_gene"
        transcript2gene[t.id] = (gene_id, gene_names.get(gene_id, gene_id))

        transcript_name = t.id
        if "transcript_name" in t.attributes:
            transcript_name = t["transcript_name"][0]
        transcript_names[t.id] = transcript_name

    logger.info("Loaded %d genes and %d transcripts" % (len(gene_names), len(transcript_names)))
    return gene_names, transcript_names, transcript2gene


def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("--output", "-o", type=str, help="output prefix name", required=True)
    parser.add_argument("--input", "-i", type=str, help="counts in linear IsoQuant format", required=True)
    parser.add_argument("--genedb", "-g", type=str, help="gene annotation in .db format "
                                                         "(can be found in IsoQuant output folder), "
                                                         "feature names will be used instead of IDs if provided")
    parser.add_argument("--feature_type", help="feature types to be converted [gene, transcript]; "
                                               "has no effect if the annotation is not provided",
                        default="gene", choices=['gene', 'transcript'])
    parser.add_argument("--output_format", "-f", type=str, choices=["mtx", "matrix"],
                        help="output format [mtx, matrix]", required=True)
    parser.add_argument("--tpm", help="convert counts to TPM", dest="convert_to_tpm",
                        action='store_true', default=False)
    parser.add_argument("--gzip", help="gzip output files", dest="gzipped",
                        action='store_true', default=False)

    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    set_logger(logger)

    feature_name_dict = None
    if args.genedb:
        gene_names, transcript_names, transcript2gene = load_gene_name_map(args.genedb)
        if args.feature_type == 'gene':
            feature_name_dict = gene_names
        elif args.feature_type == 'transcript':
            feature_name_dict = transcript_names
            with open(args.output + ".transcript2gene.tsv", "w") as out_mapf:
                for t in sorted(transcript2gene.keys()):
                    gene_info = transcript2gene[t]
                    out_mapf.write("%s\t%s\t%s\n" % (t, gene_info[0], gene_info[1]))
        else:
            logger.error("Unknown feature %s" % args.feature_type)
            sys.exit(IsoQuantExitCode.INVALID_PARAMETER)

    if args.output_format == "mtx":
        convert_to_mtx(args.input, args.output, feature_name_dict, gzipped=args.gzipped,
                       convert_to_tpm=args.convert_to_tpm)
    elif args.output_format == "matrix":
        convert_to_matrix(args.input, args.output, feature_name_dict, gzipped=args.gzipped,
                          feature_type=args.feature_type, convert_to_tpm=args.convert_to_tpm)
    else:
        logger.error("Unknown format %s" % args.output_format)
        sys.exit(IsoQuantExitCode.INVALID_PARAMETER)


if __name__ == "__main__":
   # stuff only to run when not called via 'import' here
    try:
        main()
    except SystemExit:
        raise
    except:
        print_exc()
        sys.exit(IsoQuantExitCode.UNCAUGHT_EXCEPTION)



############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2019-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import os
import sys
import logging
import yaml
from collections import defaultdict
from enum import Enum, unique

from .error_codes import IsoQuantExitCode
from .file_utils import normalize_path


logger = logging.getLogger('IsoQuant')


@unique
class InputDataType(Enum):
    undefined = 0
    fastq = 1
    bam = 2
    unmapped_bam = 3
    save = 10

    def needs_mapping(self):
        return self in [InputDataType.fastq, InputDataType.unmapped_bam]


class SampleData:
    def __init__(self, file_list, prefix, out_dir, readable_names_dict, illumina_bam, barcoded_reads=None):
        # list of lists, since each sample may contain several libraries, and each library may contain 2 files (paired)
        self.file_list = file_list
        self.readable_names_dict = readable_names_dict
        self.illumina_bam = illumina_bam
        self.prefix = prefix
        self.out_dir = out_dir
        self.aux_dir = os.path.join(self.out_dir, "aux")
        if not barcoded_reads:
            self.barcoded_reads = []
        else:
            self.barcoded_reads = barcoded_reads
        self.use_technical_replicas = False  # Will be set by DatasetProcessor
        self._init_paths()

    def _make_path(self, name):
        return os.path.join(self.out_dir, name)

    def _make_aux_path(self, name):
        return os.path.join(self.aux_dir, name)

    def _init_paths(self):
        self.out_assigned_tsv = self._make_path(self.prefix + ".read_assignments.tsv")
        self.out_assigned_tsv_result = self.out_assigned_tsv
        self.out_raw_file = self._make_aux_path(self.prefix + ".save")
        self.read_group_file = self._make_aux_path(self.prefix + ".read_group")
        self.out_corrected_bed = self._make_path(self.prefix + ".corrected_reads.bed")
        self.out_alt_tsv = self._make_path(self.prefix + ".read_assignments.SQANTI-like.tsv")
        self.out_gene_counts_tsv = self._make_path(self.prefix + ".gene")
        self.out_transcript_counts_tsv = self._make_path(self.prefix + ".transcript")
        self.out_transcript_model_counts_tsv = self._make_path(self.prefix + ".discovered_transcript")
        self.out_gene_model_counts_tsv = self._make_path(self.prefix + ".discovered_gene")
        self.out_transcript_model_grouped_counts_tsv = self._make_path(self.prefix + ".discovered_transcript_grouped")
        self.out_gene_model_grouped_counts_tsv = self._make_path(self.prefix + ".discovered_gene_grouped")
        self.out_exon_counts_tsv = self._make_path(self.prefix + ".exon")
        self.out_intron_counts_tsv = self._make_path(self.prefix + ".intron")
        self.out_gene_grouped_counts_tsv = self._make_path(self.prefix + ".gene_grouped")
        self.out_transcript_grouped_counts_tsv = self._make_path(self.prefix + ".transcript_grouped")
        self.out_exon_grouped_counts_tsv = self._make_path(self.prefix + ".exon_grouped")
        self.out_intron_grouped_counts_tsv = self._make_path(self.prefix + ".intron_grouped")
        self.out_t2t_tsv = self._make_path(self.prefix + ".novel_vs_known.SQANTI-like.tsv")
        self.barcodes_tsv = self._make_path(self.prefix + ".barcoded_reads")
        self.barcodes_done = self._make_aux_path(self.prefix + ".barcodes_done")
        self.barcodes_split_reads = self._make_aux_path(self.prefix + ".split_barcodes")
        self.out_umi_filtered = self._make_path(self.prefix + ".UMI_filtered")
        self.out_umi_filtered_tmp = self._make_aux_path(self.prefix + ".UMI_filtered")
        self.out_umi_filtered_done= self._make_aux_path(self.prefix + ".UMI_filtered.done")
        self.split_reads_fasta = self._make_path(self.prefix + ".split_reads")

    # Chromosome-specific path methods (delegate to file_naming.py)

    def get_save_file(self, chr_id: str) -> str:
        """Get path to serialized read assignments for a chromosome."""
        from .file_naming import saves_file_name
        return saves_file_name(self.out_raw_file, chr_id)

    def get_multimappers_file(self, chr_id: str) -> str:
        """Get path to multimapper assignments for a chromosome."""
        from .file_naming import multimappers_file_name
        return multimappers_file_name(self.out_raw_file, chr_id)

    def get_filtered_reads_file(self, chr_id: str) -> str:
        """Get path to filtered reads for a chromosome."""
        from .file_naming import filtered_reads_file_name
        return filtered_reads_file_name(self.out_raw_file, chr_id)

    def get_dynamic_pools_file(self, chr_id: str) -> str:
        """Get path to dynamic string pools for a chromosome."""
        from .file_naming import dynamic_pools_file_name
        return dynamic_pools_file_name(self.out_raw_file, chr_id)

    def get_barcode_pools_file(self, chr_id: str) -> str:
        """Get path to barcode/UMI pools for a chromosome (--barcoded_bam mode)."""
        from .file_naming import barcode_pools_file_name
        return barcode_pools_file_name(self.out_raw_file, chr_id)

    def get_collected_lock_file(self, chr_id: str) -> str:
        """Get path to lock file indicating reads collected for a chromosome."""
        from .file_naming import reads_collected_lock_file_name
        return reads_collected_lock_file_name(self.out_raw_file, chr_id)

    def get_processed_lock_file(self, chr_id: str) -> str:
        """Get path to lock file indicating reads processed for a chromosome."""
        from .file_naming import reads_processed_lock_file_name
        return reads_processed_lock_file_name(self.out_raw_file, chr_id)

    def get_read_group_split_file(self, chr_id: str, spec_index: int = None) -> str:
        """Get path to split read group file for a chromosome.

        Args:
            chr_id: Chromosome ID
            spec_index: Optional spec index for multi-spec files

        Returns:
            Path like 'prefix.read_group_spec0_chr1' or 'prefix.read_group_chr1'
        """
        from .file_naming import convert_chr_id_to_file_name_str
        chr_str = convert_chr_id_to_file_name_str(chr_id)
        if spec_index is not None:
            return f"{self.read_group_file}_spec{spec_index}_{chr_str}"
        return f"{self.read_group_file}_{chr_str}"

    def get_barcodes_split_file(self, chr_id: str) -> str:
        """Get path to split barcodes file for a chromosome."""
        from .file_naming import convert_chr_id_to_file_name_str
        return self.barcodes_split_reads + "_" + convert_chr_id_to_file_name_str(chr_id)

    # Chromosome-specific output file getters (for parallel processing)

    def get_chr_prefix(self, chr_id: str) -> str:
        """Get chromosome-specific prefix for output files."""
        from .file_naming import convert_chr_id_to_file_name_str
        return f"{self.prefix}_{convert_chr_id_to_file_name_str(chr_id)}"

    def get_corrected_bed_file(self, chr_id: str) -> str:
        """Get path to corrected reads BED file for a chromosome."""
        return self._make_path(self.get_chr_prefix(chr_id) + ".corrected_reads.bed")

    def get_assigned_tsv_file(self, chr_id: str) -> str:
        """Get path to read assignments TSV for a chromosome."""
        return self._make_path(self.get_chr_prefix(chr_id) + ".read_assignments.tsv")

    def get_gene_counts_file(self, chr_id: str) -> str:
        """Get path to gene counts file for a chromosome."""
        return self._make_path(self.get_chr_prefix(chr_id) + ".gene")

    def get_transcript_counts_file(self, chr_id: str) -> str:
        """Get path to transcript counts file for a chromosome."""
        return self._make_path(self.get_chr_prefix(chr_id) + ".transcript")

    def get_exon_counts_file(self, chr_id: str) -> str:
        """Get path to exon counts file for a chromosome."""
        return self._make_path(self.get_chr_prefix(chr_id) + ".exon")

    def get_intron_counts_file(self, chr_id: str) -> str:
        """Get path to intron counts file for a chromosome."""
        return self._make_path(self.get_chr_prefix(chr_id) + ".intron")

    def get_transcript_model_counts_file(self, chr_id: str) -> str:
        """Get path to discovered transcript counts file for a chromosome."""
        return self._make_path(self.get_chr_prefix(chr_id) + ".discovered_transcript")

    def get_gene_model_counts_file(self, chr_id: str) -> str:
        """Get path to discovered gene counts file for a chromosome."""
        return self._make_path(self.get_chr_prefix(chr_id) + ".discovered_gene")

    def get_t2t_tsv_file(self, chr_id: str) -> str:
        """Get path to SQANTI-like TSV for a chromosome."""
        return self._make_path(self.get_chr_prefix(chr_id) + ".novel_vs_known.SQANTI-like.tsv")

    def get_grouped_counts_file(self, chr_id: str, feature: str, strategy_name: str) -> str:
        """Get path to grouped counts file for a chromosome.

        Args:
            chr_id: Chromosome ID
            feature: Feature type (gene, transcript, exon, intron, discovered_transcript, discovered_gene)
            strategy_name: Grouping strategy name
        """
        return self._make_path(f"{self.get_chr_prefix(chr_id)}.{feature}_grouped_{strategy_name}")

    # Sample-level auxiliary files (for resume functionality)

    def get_info_file(self) -> str:
        """Get path to sample collection info file."""
        from .file_naming import info_file_name
        return info_file_name(self.out_raw_file)

    def get_collection_lock_file(self) -> str:
        """Get path to sample collection lock file."""
        from .file_naming import collection_lock_file_name
        return collection_lock_file_name(self.out_raw_file)


class InputDataStorage:
    def __init__(self, args):
        # list of SampleData
        self.samples = []
        self.input_type = InputDataType.undefined
        readable_names_dict = defaultdict(lambda: defaultdict(str))
        sample_files = []
        experiment_names = []
        self.experiment_prefix = args.prefix
        illumina_bam = []

        if args.fastq is not None:
            self.input_type = InputDataType.fastq
            sample_files.append([])
            experiment_name = args.prefix
            experiment_names.append(experiment_name)
            if args.labels and len(args.labels) != len(args.fastq):
                logger.critical("Number of labels is not equal to the number of files")
                sys.exit(IsoQuantExitCode.LABEL_COUNT_MISMATCH)
            for i, fq in enumerate(args.fastq):
                check_input_type(fq, self.input_type)
                if fq in readable_names_dict[experiment_name]:
                    logger.critical("File %s is used multiple times in a single experiment, which is not allowed" %fq)
                    sys.exit(IsoQuantExitCode.DUPLICATE_FILES)
                sample_files[0].append([fq])
                readable_names_dict[experiment_name][fq] = args.labels[i] if args.labels else \
                    os.path.splitext(os.path.basename(fq))[0]
            # if args.illumina_bam is not None:
                # illumina_bam.append(args.illumina_bam)
            # else:
                # illumina_bam.append([None])
            illumina_bam.append(args.illumina_bam)

        elif args.bam is not None:
            self.input_type = InputDataType.bam
            sample_files.append([])
            experiment_name = args.prefix
            experiment_names.append(experiment_name)
            if args.labels and len(args.labels) != len(args.bam):
                logger.critical("Number of labels is not equal to the number of files")
                sys.exit(IsoQuantExitCode.LABEL_COUNT_MISMATCH)
            for i, bam in enumerate(args.bam):
                check_input_type(bam, self.input_type)
                if bam in readable_names_dict[experiment_name]:
                    logger.critical("File %s is used multiple times in a single experiment, which is not allowed" %bam)
                    sys.exit(IsoQuantExitCode.DUPLICATE_FILES)
                sample_files[0].append([bam])
                readable_names_dict[experiment_name][bam] = args.labels[i] if args.labels else \
                    os.path.splitext(os.path.basename(bam))[0]
            illumina_bam.append(args.illumina_bam)

        elif args.unmapped_bam is not None:
            self.input_type = InputDataType.unmapped_bam
            sample_files.append([])
            experiment_name = args.prefix
            experiment_names.append(experiment_name)
            if args.labels and len(args.labels) != len(args.unmapped_bam):
                logger.critical("Number of labels is not equal to the number of files")
                sys.exit(IsoQuantExitCode.LABEL_COUNT_MISMATCH)
            for i, bam in enumerate(args.unmapped_bam):
                check_input_type(bam, self.input_type)
                if bam in readable_names_dict[experiment_name]:
                    logger.critical("File %s is used multiple times in a single experiment, which is not allowed" %bam)
                    sys.exit(IsoQuantExitCode.DUPLICATE_FILES)
                sample_files[0].append([bam])
                readable_names_dict[experiment_name][bam] = args.labels[i] if args.labels else \
                    os.path.splitext(os.path.basename(bam))[0]
            illumina_bam.append(args.illumina_bam)

        elif args.read_assignments is not None:
            self.input_type = InputDataType.save
            illumina_bam = [[]]
            for i, save_file in enumerate(args.read_assignments):
                sample_files.append([[save_file]])
                experiment_names.append(self.experiment_prefix + str(i))
        
        elif args.yaml is not None:
            sample_files, experiment_names, readable_names_dict, illumina_bam = self.get_samples_from_yaml(args.yaml)
            if args.labels:
                logger.warning("--labels option has no effect when files are provided via yaml file")

        else:
            logger.critical("Input data was not specified")
            sys.exit(IsoQuantExitCode.NO_INPUT_DATA)

        for i in range(len(sample_files)):
            self.samples.append(SampleData(sample_files[i], experiment_names[i],
                                           os.path.join(args.output, experiment_names[i]),
                                           readable_names_dict[experiment_names[i]],
                                           illumina_bam[i]))

    def get_samples_from_file(self, file_name):
        sample_files = []
        experiment_names = []
        illumina_bam = []
        readable_names_dict = defaultdict(lambda: defaultdict(str))
        inf = open(file_name, "r")
        current_sample = []
        current_sample_name = self.experiment_prefix
        current_index = 0

        for l in inf:
            if len(l.strip()) == 0 or l.startswith("#"):
                if len(current_sample) > 0:
                    sample_files.append(current_sample)
                    experiment_names.append(current_sample_name)
                    illumina_bam.append(None)
                current_sample = []
                current_sample_name = l.strip()[1:]
                if not current_sample_name:
                    current_sample_name = self.experiment_prefix + str(current_index)
                if current_sample_name in experiment_names:
                    new_sample_name = self.experiment_prefix + str(current_index)
                    if current_sample_name == new_sample_name:
                        logger.critical("Change experiment name %s and rerun IsoQuant" % current_sample_name)
                        sys.exit(IsoQuantExitCode.DUPLICATE_FILES)
                    logger.warning("Duplicate folder prefix %s, will change to %s" %
                                   (current_sample_name, new_sample_name))
                    current_sample_name = new_sample_name
                current_index += 1
            else:
                vals = l.strip().split(':')
                files = vals[0].split()
                if len(vals) > 1:
                    readable_name = vals[-1]
                else:
                    readable_name = os.path.splitext(os.path.basename(files[0]))[0]
                current_sample.append(files)
                for fname in files:
                    if fname in readable_names_dict[current_sample_name]:
                        logger.critical("File %s is used multiple times in a single experiment, which is not allowed" %fname)
                        sys.exit(IsoQuantExitCode.DUPLICATE_FILES)
                    readable_names_dict[current_sample_name][fname] = readable_name

        if len(current_sample) > 0:
            sample_files.append(current_sample)
            experiment_names.append(current_sample_name)
            illumina_bam.append(None)

        for sample in sample_files:
            for lib in sample:
                for in_file in lib:
                    check_input_type(in_file, self.input_type)

        return sample_files, experiment_names, readable_names_dict, illumina_bam

    def has_replicas(self):
        return any(len(sample.file_list) > 1 for sample in self.samples)
        
    def get_samples_from_yaml(self, yaml_file_path):
        sample_files = []
        experiment_names = []
        illumina_bam = []
        readable_names_dict = defaultdict(lambda: defaultdict(str))
        yaml_file = open(yaml_file_path, 'r')
        con = yaml.safe_load(yaml_file)
        current_index = 0
        t = con[0]
        if not 'data format' in t.keys():
            logger.critical("Please specify whether you are using fastq or bam files in the first entry")
            sys.exit(IsoQuantExitCode.YAML_PARSING_ERROR)
        else:
            if len(t.keys()) > 1:
                logger.warning("The first entry should only specify the input data format. Any additional info will be ignored")
            if  t['data format'] == "bam":
                self.input_type = InputDataType.bam
            elif t['data format'] == "unmapped_bam":
                self.input_type = InputDataType.unmapped_bam
            elif t['data format'] == "fastq" or t['data format'] == "fasta":
                self.input_type = InputDataType.fastq
            else:
                logger.critical("The input data format can only be either fastq, fasta or bam.")
                sys.exit(IsoQuantExitCode.INVALID_FILE_FORMAT)
        for sample in con[1:]:
            if not 'name' in sample.keys():
                current_sample_name = self.experiment_prefix + str(current_index)
            else:
                current_sample_name = sample['name']
            if current_sample_name in experiment_names:
                    new_sample_name = self.experiment_prefix + str(current_index)
                    if current_sample_name == new_sample_name:
                        logger.critical("Change experiment name %s and rerun IsoQuant" % current_sample_name)
                        sys.exit(IsoQuantExitCode.DUPLICATE_FILES)
                    logger.warning("Duplicate folder prefix %s, will change to %s" %
                                   (current_sample_name, new_sample_name))
                    current_sample_name = new_sample_name
            current_index += 1
            if not 'long read files' in sample.keys():
                logger.critical("Experiment %s does not contain any files" %current_sample_name)
                sys.exit(IsoQuantExitCode.YAML_PARSING_ERROR)
            else:
                current_sample = [normalize_path(yaml_file_path, b) for b in sample['long read files']]
                names = 'labels' in sample.keys()
                if names and not len(sample['labels']) == len(current_sample):
                    logger.critical("The number of file aliases differs from the number of files")
                    sys.exit(IsoQuantExitCode.LABEL_COUNT_MISMATCH)
                for f in range(len(current_sample)):
                    fname = current_sample[f]
                    if names:
                        readable_name = sample['labels'][f]
                    else:
                        readable_name = os.path.splitext(os.path.basename(fname))[0]
                    if fname in readable_names_dict[current_sample_name]:
                        logger.critical("File %s is used multiple times in a single experiment, which is not allowed" %fname)
                        sys.exit(IsoQuantExitCode.DUPLICATE_FILES)
                    readable_names_dict[current_sample_name][fname] = readable_name
            if len(current_sample) > 0:
                current_sample_list = [[s] for s in current_sample]
                sample_files.append(current_sample_list)
                experiment_names.append(current_sample_name)
                if 'illumina bam' in sample.keys():
                    illumina_bam.append([normalize_path(yaml_file_path, ib) for ib in sample['illumina bam']])
                else:
                    illumina_bam.append(None)
            
        for sample in sample_files:
            for lib in sample:
                for in_file in lib:
                    check_input_type(in_file, self.input_type)
        return sample_files, experiment_names, readable_names_dict, illumina_bam
        
# not functional yet
# idea for the future to name unnamed samples by their last common folder
    # def get_sample_name(names, index):
        # common_characters = len(names[0])
        # common_name = names[0]
        
        # for i in range(1, len(names)):
            # p = mismatch(common_name, names[i])
            # if p[0] < common_characters:
                # common_characters = p[0]
                
        # found = common_names.rfind('/', 0, common_characters)
        
        # common_name = common_name[:found]
        
        # sample_name = common_name + str(index)
        # return sample_name


def check_input_type(fname, input_type):
    basename_plus_inner_ext, outer_ext = os.path.splitext(fname.lower())
    if outer_ext not in ['.zip', '.gz', '.gzip', '.bz2', '.bzip2']:
        basename_plus_inner_ext, outer_ext = fname, ''  # not a supported archive

    basename, fasta_ext = os.path.splitext(basename_plus_inner_ext)
    if fasta_ext in ['.fastq', '.fasta', '.fa', '.fq', '.fna']:
        if input_type != InputDataType.fastq:
            raise Exception("Wrong file extension was detected %s. Use only FASTQ/FASTA files with --fastq option." % fname)
    elif fasta_ext == '.bam':
        if input_type not in [InputDataType.bam, InputDataType.unmapped_bam]:
            raise Exception("Wrong file extension was detected for file %s. Use only BAM files with --bam option." % fname)
    else:
        raise Exception("File format " + fasta_ext + " is not supported! Supported formats: FASTQ, FASTA, BAM")




############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2020-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging

from .common import (
    contains,
    contains_approx,
    contains_well_inside,
    equal_ranges,
    get_exon,
    get_following_exon_from_junctions,
    get_preceding_exon_from_junctions,
    interval_len,
    left_of,
    overlaps,
    overlaps_at_least,
)
from .isoform_assignment import (
    MatchEvent,
    MatchEventSubtype,
    SupplementaryMatchConstants,
    alternative_sites
)


logger = logging.getLogger('IsoQuant')


class JunctionComparator:
    def __init__(self, params, intron_profile_constructor):
        self.params = params
        self.intron_profile_constructor = intron_profile_constructor

    def compare_junctions(self, read_junctions, read_region, isoform_junctions, isoform_region):
        """ compare read splice junctions against similar isoform

        Parameters
        ----------
        read_junctions: list of tuple of int
        read_region: tuple of int
        isoform_junctions: list of tuple of int
        isoform_region: tuple of int

        Returns
        -------
        list of detected contradiction events
        """
        if not read_junctions:
            return self.get_mono_exon_subtype(read_region, isoform_junctions)

        read_pos = 0
        isoform_pos = 0
        read_features_present = [0 for i in range(0, len(read_junctions))]
        isoform_features_present = [0 for i in range(0, len(isoform_junctions))]
        contradictory_region_pairs = []
        absent_region = (SupplementaryMatchConstants.absent_position, SupplementaryMatchConstants.absent_position)
        current_contradictory_region = absent_region

        while read_pos < len(read_junctions) and isoform_pos < len(isoform_junctions):
            if equal_ranges(isoform_junctions[isoform_pos], read_junctions[read_pos], self.params.delta):
                # junctions are equal
                read_features_present[read_pos] = 1
                isoform_features_present[isoform_pos] = 1
                if current_contradictory_region != absent_region:
                    contradictory_region_pairs.append(current_contradictory_region)
                    current_contradictory_region = absent_region
                read_pos += 1
                isoform_pos += 1

            elif overlaps(isoform_junctions[isoform_pos], read_junctions[read_pos]):
                # junctions overlap, but are unequal
                read_features_present[read_pos] = -1
                isoform_features_present[isoform_pos] = -1
                if current_contradictory_region == absent_region:
                    current_contradictory_region = ((read_pos, read_pos), (isoform_pos, isoform_pos))
                else:
                    current_contradictory_region = (
                    (current_contradictory_region[0][0], read_pos), (current_contradictory_region[1][0], isoform_pos))
                if read_junctions[read_pos][1] < isoform_junctions[isoform_pos][1]:
                    read_pos += 1
                else:
                    isoform_pos += 1

            elif left_of(isoform_junctions[isoform_pos], read_junctions[read_pos]):
                # isoform junction is behind, move on
                if current_contradictory_region != absent_region:
                    contradictory_region_pairs.append(current_contradictory_region)
                    current_contradictory_region = absent_region
                if read_pos > 0 or overlaps(read_region, isoform_junctions[isoform_pos]):
                    if isoform_features_present[isoform_pos] != -1:
                        contradictory_region_pairs.append(((SupplementaryMatchConstants.absent_position, read_pos), (isoform_pos, isoform_pos)))
                    isoform_features_present[isoform_pos] = -1
                isoform_pos += 1

            else:
                # read junction is behind, move on
                if current_contradictory_region != absent_region:
                    contradictory_region_pairs.append(current_contradictory_region)
                    current_contradictory_region = absent_region
                if isoform_pos > 0 or overlaps(isoform_region, read_junctions[read_pos]):
                    if read_features_present[read_pos] != -1:
                        contradictory_region_pairs.append(((read_pos, read_pos), (SupplementaryMatchConstants.absent_position, isoform_pos)))
                    read_features_present[read_pos] = -1
                read_pos += 1

        if current_contradictory_region != absent_region:
            contradictory_region_pairs.append(current_contradictory_region)

        # check terminating regions
        while read_pos < len(read_junctions):
            if overlaps(isoform_region, read_junctions[read_pos]):
                if read_features_present[read_pos] != -1:
                    contradictory_region_pairs.append(((read_pos, read_pos), (SupplementaryMatchConstants.absent_position, isoform_pos)))
                    read_features_present[read_pos] = -1
            else:
                break
            read_pos += 1

        while isoform_pos < len(isoform_junctions):
            if overlaps(read_region, isoform_junctions[isoform_pos]):
                if isoform_features_present[isoform_pos] != -1:
                    contradictory_region_pairs.append(((SupplementaryMatchConstants.absent_position, read_pos), (isoform_pos, isoform_pos)))
                    isoform_features_present[isoform_pos] = -1
            else:
                break
            isoform_pos += 1

        # logger.debug("+ + Inspected contradictory read")
        # logger.debug("+ + Read profile " + str(read_features_present))
        # logger.debug("+ + Read introns " + str(read_junctions))
        # logger.debug("+ + Read region " + str(read_region))

        # logger.debug("+ + Isoform profile " + str(isoform_features_present))
        # logger.debug("+ + Isoform introns " + str(isoform_junctions))
        # logger.debug("+ + Isoform region " + str(isoform_region))

        matching_events = []
        # logger.debug(read_features_present)
        # logger.debug(isoform_features_present)
        if any(el == -1 for el in read_features_present) or any(el == -1 for el in isoform_features_present):
            # classify contradictions
            # logger.debug("+ + Classifying contradictions")
            matching_events = self.detect_contradiction_type(read_region, read_junctions, isoform_region, isoform_junctions, contradictory_region_pairs)

        if read_features_present[0] == 0 or read_features_present[-1] == 0:
            # if all(x == 0 for x in read_features_present):
            #    return [MatchEvent(MatchEventSubtype.undefined)]
            # logger.debug("+ + Found only extra terminal introns ")
            self.add_extra_out_exon_events(matching_events, read_features_present, read_region, read_junctions, isoform_region[0])

        if len(matching_events) == 0:
            # logger.debug("No contradiction detected")
            return [MatchEvent(MatchEventSubtype.none)]
        return matching_events

    def detect_contradiction_type(self, read_region, read_junctions, isoform_region, isoform_junctions, contradictory_region_pairs):
        """

        Parameters
        ----------
        read_junctions: list of tuples of int
        isoform_junctions: list of tuples of int
        contradictory_region_pairs: list of tuples of int

        Returns
        -------
        list of contradiction events
        """
        contradiction_events = []
        for pair in contradictory_region_pairs:
            # classify each contradictory area separately
            event = self.compare_overlapping_contradictional_regions(read_region, read_junctions, isoform_region, isoform_junctions, pair[0], pair[1])
            if event is not None:
                contradiction_events.append(event)

        return contradiction_events

    def compare_overlapping_contradictional_regions(self, read_region, read_junctions,
                                                    isoform_region, isoform_junctions,
                                                    read_cregion, isoform_cregion):
        if read_cregion[0] == SupplementaryMatchConstants.absent_position:
            if isoform_cregion[0] != isoform_cregion[1]:
                logger.warning("Multiple intron retentions in a single event:" + str(isoform_cregion))
            overlapped_isoform_intron = isoform_junctions[isoform_cregion[0]]
            if contains(read_region, overlapped_isoform_intron):
                if interval_len(overlapped_isoform_intron) <= self.params.micro_intron_length and \
                        contains_well_inside(
                            get_preceding_exon_from_junctions(read_region, read_junctions, read_cregion[1]),
                            overlapped_isoform_intron, self.params.minimal_exon_overlap):
                    return MatchEvent(MatchEventSubtype.fake_micro_intron_retention, isoform_cregion, read_cregion)
                return MatchEvent(MatchEventSubtype.intron_retention, isoform_cregion, read_cregion)
            elif overlaps_at_least(read_region, overlapped_isoform_intron, self.params.minor_exon_extension):
                if overlapped_isoform_intron[0] <= read_region[0]:
                    return MatchEvent(MatchEventSubtype.incomplete_intron_retention_left, isoform_cregion, read_cregion)
                else:
                    return MatchEvent(MatchEventSubtype.incomplete_intron_retention_right, isoform_cregion, read_cregion)
            else:
                return None

        elif isoform_cregion[0] == SupplementaryMatchConstants.absent_position:
            if self.are_known_introns(read_junctions, read_cregion):
                return MatchEvent(MatchEventSubtype.extra_intron_known, isoform_cregion, read_cregion)
            elif self.are_suspicious_introns(read_region, read_junctions, read_cregion):
                return MatchEvent(MatchEventSubtype.none, isoform_cregion, read_cregion)
            elif read_cregion[0] == 0 and \
                    interval_len(get_exon(read_region, read_junctions, 0)) <= self.params.max_fake_terminal_exon_len:
                # we have extra intron to the left and first exons is short
                return MatchEvent(MatchEventSubtype.fake_terminal_exon_left,
                                  SupplementaryMatchConstants.extra_left_region, read_cregion)
            elif read_cregion[1] == len(read_junctions) - 1 and \
                    interval_len(get_exon(read_region, read_junctions, -1)) <= self.params.max_fake_terminal_exon_len:
                # we have extra intron to the right and last exons is short
                return MatchEvent(MatchEventSubtype.fake_terminal_exon_right,
                                  SupplementaryMatchConstants.extra_right_region, read_cregion)
            return MatchEvent(MatchEventSubtype.extra_intron_novel, isoform_cregion, read_cregion)

        read_intron_total_len = sum(interval_len(read_junctions[i])
                                    for i in range(read_cregion[0], read_cregion[1] + 1))
        isoform_intron_total_len = sum(interval_len(isoform_junctions[i])
                                       for i in range(isoform_cregion[0], isoform_cregion[1] + 1))
        total_intron_len_diff = abs(read_intron_total_len - isoform_intron_total_len)
        intron_length_is_similar = \
            total_intron_len_diff <= min(self.params.max_intron_abs_diff,
                                         self.params.max_intron_rel_diff * max(read_intron_total_len, isoform_intron_total_len))

        read_introns_known = self.are_known_introns(read_junctions, read_cregion)

        left_exons_overlap = overlaps(get_exon(read_region, read_junctions, read_cregion[0]),
                                      get_exon(isoform_region, isoform_junctions, isoform_cregion[0]))
        right_exons_overlap = overlaps(get_exon(read_region, read_junctions, read_cregion[1] + 1),
                                       get_exon(isoform_region, isoform_junctions, isoform_cregion[1] + 1))
        surrounded_by_exons = left_exons_overlap and right_exons_overlap
        read_left_site = read_junctions[read_cregion[0]][0]
        read_right_site = read_junctions[read_cregion[1]][1]
        isoform_left_site = isoform_junctions[isoform_cregion[0]][0]
        isoform_right_site = isoform_junctions[isoform_cregion[1]][1]
        similar_letf_site = abs(read_left_site - isoform_left_site) <= 2 * self.params.delta
        similar_right_site = abs(read_right_site - isoform_right_site) <= 2 * self.params.delta
        similar_bounds = similar_letf_site and similar_right_site
        read_introns_inside = contains_approx((isoform_left_site, isoform_right_site),
                                              (read_left_site, read_right_site), self.params.delta)
        isofrom_introns_inside = contains_approx((read_left_site, read_right_site),
                                                 (isoform_left_site, isoform_right_site), self.params.delta)

        event = None
        if surrounded_by_exons and read_cregion[1] == read_cregion[0] and isoform_cregion[1] == isoform_cregion[0]:
            # single exons is altered
            event = self.classify_single_intron_alternation(read_region, read_junctions, isoform_region,
                                                            isoform_junctions, read_cregion[0], isoform_cregion[0],
                                                            intron_length_is_similar, read_introns_known)

        elif len(read_junctions) > 1 and \
                read_cregion[1] == read_cregion[0] and isoform_cregion[1] == isoform_cregion[0] and \
                ((read_cregion[0] == 0 and isoform_cregion[0] == 0 and similar_right_site) or
                 (read_cregion[0] == len(read_junctions)-1 and isoform_cregion[0] == len(isoform_junctions) - 1
                  and similar_letf_site)):
            # terminal exon alternation
            if read_cregion[0] == 0 and isoform_cregion[0] == 0:
                read_exon = get_preceding_exon_from_junctions(read_region, read_junctions, 0)
                isoform_exon = get_preceding_exon_from_junctions(isoform_region, isoform_junctions, 0)
            else:
                read_exon = get_following_exon_from_junctions(read_region, read_junctions, -1)
                isoform_exon = get_following_exon_from_junctions(isoform_region, isoform_junctions, -1)
            if abs(interval_len(read_exon) - interval_len(isoform_exon)) < 2 * self.params.delta:
                # TODO: verify with alignment
                if read_cregion[0] == 0:
                    event = MatchEventSubtype.terminal_exon_misalignment_left
                else:
                    event = MatchEventSubtype.terminal_exon_misalignment_right
            elif read_introns_known:
                event = MatchEventSubtype.terminal_exon_shift_known
            else:
                event = MatchEventSubtype.terminal_exon_shift_novel

        elif surrounded_by_exons and similar_bounds and \
                read_cregion[1] - read_cregion[0] == isoform_cregion[1] - isoform_cregion[0] >= 1 and \
                total_intron_len_diff <= 2 * self.params.delta:
                # FIXME: increase threshold when re-alignment is implemented
            # several introns of the same total length as reference onces
            if read_introns_known:
                event = MatchEventSubtype.mutually_exclusive_exons_known
            else:
                event = MatchEventSubtype.mutually_exclusive_exons_novel

        elif surrounded_by_exons and read_introns_inside and \
                read_cregion[1] == read_cregion[0] and isoform_cregion[1] > isoform_cregion[0]:
            # skipped exon(s)
            event = self.classify_skipped_exons(isoform_junctions, isoform_cregion,
                                                intron_length_is_similar, read_introns_known, similar_bounds)

        elif surrounded_by_exons and similar_bounds and \
                read_cregion[1] > read_cregion[0] and isoform_cregion[1] == isoform_cregion[0]:
            # gain exon
            if read_introns_known:
                event = MatchEventSubtype.exon_gain_known
            elif self.are_suspicious_introns(read_region, read_junctions, read_cregion):
                event = MatchEventSubtype.intron_retention
            else:
                event = MatchEventSubtype.exon_gain_novel

        elif surrounded_by_exons and intron_length_is_similar and isofrom_introns_inside and \
                read_cregion[1] > read_cregion[0] and isoform_cregion[1] == isoform_cregion[0]:
            # exon detach
            if read_introns_known:
                event = MatchEventSubtype.exon_detach_known
            else:
                event = MatchEventSubtype.exon_detach_novel

        if event is None:
            # none of above, complex alternative structure
            if read_introns_known:
                event = MatchEventSubtype.alternative_structure_known
            elif surrounded_by_exons and self.are_suspicious_introns(read_region, read_junctions, read_cregion):
                event = MatchEventSubtype.intron_retention
            else:
                event = MatchEventSubtype.alternative_structure_novel
        return MatchEvent(event, isoform_cregion, read_cregion)

    def classify_skipped_exons(self, isoform_junctions, isoform_cregion,
                               intron_length_is_similar, read_introns_known, similar_bounds):
        total_exon_len = sum([isoform_junctions[i + 1][0] - isoform_junctions[i][1] + 1
                              for i in range(isoform_cregion[0], isoform_cregion[1])])
        event = None
        if intron_length_is_similar:
            if total_exon_len <= self.params.max_missed_exon_len:
                event = MatchEventSubtype.exon_misalignment
            elif read_introns_known:
                event = MatchEventSubtype.exon_merge_known
            else:
                event = MatchEventSubtype.exon_merge_novel
        elif similar_bounds:
            if read_introns_known:
                event = MatchEventSubtype.exon_skipping_known
            else:
                event = MatchEventSubtype.exon_skipping_novel
        return event

    def classify_single_intron_alternation(self, read_region, read_junctions, isoform_region, isoform_junctions,
                                           read_cpos, isoform_cpos, intron_length_is_similar, read_introns_known):
        if intron_length_is_similar:
            if abs(isoform_junctions[isoform_cpos][0] - read_junctions[read_cpos][0]) <= self.params.max_intron_shift:
                event = MatchEventSubtype.intron_shift
            elif read_introns_known:
                event = MatchEventSubtype.intron_migration
            else:
                event = MatchEventSubtype.intron_alternation_novel
        else:
            event = MatchEventSubtype.intron_alternation_known if read_introns_known \
                else MatchEventSubtype.intron_alternation_novel

            if abs(isoform_junctions[isoform_cpos][0] - read_junctions[read_cpos][0]) <= self.params.delta:
                # left splice sites are similar, check for exon overlap
                following_read_exon = get_following_exon_from_junctions(read_region, read_junctions, read_cpos)
                following_isoform_exon = get_following_exon_from_junctions(isoform_region, isoform_junctions, isoform_cpos)
                min_overlap = max(1, min(self.params.min_abs_exon_overlap,
                                         round(self.params.min_rel_exon_overlap * interval_len(following_isoform_exon))))
                if overlaps_at_least(following_read_exon, following_isoform_exon, min_overlap):
                    event = alternative_sites[("right", read_introns_known)]
            elif abs(isoform_junctions[isoform_cpos][1] - read_junctions[read_cpos][1]) <= self.params.delta:
                # right splice sites are similar, check for exon overlap
                preceding_read_exon = get_preceding_exon_from_junctions(read_region, read_junctions, read_cpos)
                preceding_isoform_exon = get_preceding_exon_from_junctions(isoform_region, isoform_junctions, isoform_cpos)
                min_overlap = max(1, min(self.params.min_abs_exon_overlap,
                                      round(self.params.min_rel_exon_overlap * interval_len(preceding_isoform_exon))))
                if overlaps_at_least(preceding_read_exon, preceding_isoform_exon, min_overlap):
                    event = alternative_sites[("left", read_introns_known)]

            if event in {MatchEventSubtype.intron_alternation_novel,
                         MatchEventSubtype.alt_left_site_novel,
                         MatchEventSubtype.alt_right_site_novel} and \
                self.are_suspicious_introns(read_region, read_junctions, (read_cpos, read_cpos)):
                event = MatchEventSubtype.intron_retention

        return event

    def are_suspicious_introns(self, read_region, read_junctions, read_cregion):
        total_intron_len = 0
        for cpos in range(read_cregion[0], read_cregion[1] + 1):
            if interval_len(read_junctions[cpos]) > self.params.max_suspicious_intron_abs_len:
                return False
            total_intron_len += interval_len(read_junctions[cpos])

        total_exon_len = 0
        for cpos in range(read_cregion[0], read_cregion[1] + 1):
            total_exon_len += interval_len(get_preceding_exon_from_junctions(read_region, read_junctions, cpos))
        total_exon_len += interval_len(get_following_exon_from_junctions(read_region, read_junctions, read_cregion[-1]))

        return float(total_intron_len) <= float(total_exon_len) * self.params.max_suspicious_intron_rel_len

    def get_mono_exon_subtype(self, read_region, isoform_junctions):
        if len(isoform_junctions) == 0:
            # both isoform and read are monoexon
            events = [MatchEvent(MatchEventSubtype.mono_exon_match)]
        else:
            # check if read captures some introns
            events = []
            for i, intron in enumerate(isoform_junctions):
                if contains(read_region, intron):
                    # full intron retention
                    if interval_len(intron) <= self.params.micro_intron_length and \
                            contains_well_inside(read_region, intron, self.params.minimal_exon_overlap):
                        events.append(MatchEvent(MatchEventSubtype.fake_micro_intron_retention, isoform_region=(i, i),
                                                 read_region=(SupplementaryMatchConstants.absent_position, 0)))
                    else:
                        events.append(MatchEvent(MatchEventSubtype.unspliced_intron_retention, isoform_region=(i, i),
                                                 read_region=(SupplementaryMatchConstants.absent_position, 0)))
                elif overlaps_at_least(read_region, intron, self.params.minor_exon_extension) and \
                        not contains(intron, read_region):
                    # partial IR
                    if intron[0] <= read_region[0]:
                        events.append(MatchEvent(MatchEventSubtype.incomplete_intron_retention_left, isoform_region=(i, i),
                                                 read_region=(SupplementaryMatchConstants.absent_position, 0)))
                    else:
                        events.append(MatchEvent(MatchEventSubtype.incomplete_intron_retention_right, isoform_region=(i, i),
                                                 read_region=(SupplementaryMatchConstants.absent_position, 0)))

        if not events:
            # monoexonic read without significant intron overlap
            events = [MatchEvent(MatchEventSubtype.mono_exonic)]
        return events

    def add_extra_out_exon_events(self, match_events, read_intron_read_profile, read_region, read_introns, isoform_start):
        extra_left = read_intron_read_profile[0] == 0
        extra_right = read_intron_read_profile[-1] == 0

        if all(x == 0 for x in read_intron_read_profile):
            if read_introns[0][0] < isoform_start:
                extra_right = False
            else:
                extra_left = False

        if extra_left:
            read_pos = 0
            if interval_len(get_exon(read_region, read_introns, read_pos)) <= self.params.max_fake_terminal_exon_len:
                match_events.append(MatchEvent(MatchEventSubtype.fake_terminal_exon_left,
                                               SupplementaryMatchConstants.extra_left_region, (read_pos, read_pos)))
                read_pos += 1

            while read_pos < len(read_intron_read_profile) and read_intron_read_profile[read_pos] == 0:
                match_events.append(MatchEvent(MatchEventSubtype.extra_intron_flanking_left,
                                               SupplementaryMatchConstants.extra_left_region, (read_pos, read_pos)))
                read_pos += 1

        if extra_right:
            read_pos = len(read_intron_read_profile) - 1
            if interval_len(get_exon(read_region, read_introns, read_pos + 1)) <= self.params.max_fake_terminal_exon_len:
                match_events.append(MatchEvent(MatchEventSubtype.fake_terminal_exon_right,
                                               SupplementaryMatchConstants.extra_right_region, (read_pos, read_pos)))
                read_pos -= 1

            while read_pos >= 0 and read_intron_read_profile[read_pos] == 0:
                match_events.append(MatchEvent(MatchEventSubtype.extra_intron_flanking_right,
                                               SupplementaryMatchConstants.extra_right_region, (read_pos, read_pos)))
                read_pos -= 1

    def profile_for_junctions_introns(self, junctions, region):
        selected_junctions = []
        # logger.debug("Checking for known introns " + str(region))
        for i in range(region[0], region[1] + 1):
            selected_junctions.append(junctions[i])

        selected_junctions_profile = self.intron_profile_constructor.construct_profile_for_features(selected_junctions)
        return selected_junctions_profile

    def are_known_introns(self, junctions, region):
        selected_junctions_profile = self.profile_for_junctions_introns(junctions, region)
        return all(el == 1 for el in selected_junctions_profile.read_profile)



############################################################################
# Copyright (c) 2025-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

import os
import pytest
import tempfile

from src.read_groups import (
    parse_barcode2spot_spec,
    load_barcode2barcode_mapping,
    BarcodeSpotGrouper,
    SharedTableData,
)
from src.barcode_calling.umi_filtering import UMIFilter
from src.isoform_assignment import ReadAssignment, ReadAssignmentType, IsoformMatch, MatchClassification
from src.common import junctions_from_blocks
from src.string_pools import StringPoolManager
from src.file_naming import umi_barcode2barcode_prefix, umi_barcode2barcode_global_lock


# -- Helpers / Mocks --

class MockAlignment:
    def __init__(self, read_id=""):
        self.query_name = read_id


class MockReadAssignment:
    """Lightweight mock used only for grouper tests (no string_pools needed)."""
    def __init__(self, barcode=None):
        self.barcode = barcode


def _make_read_assignment(string_pools, read_id, barcode, umi, gene_id, transcript_id,
                          exons, assignment_type=ReadAssignmentType.unique):
    """Create a fully populated ReadAssignment for UMI filtering tests."""
    ra = ReadAssignment(read_id=read_id, assignment_type=assignment_type, string_pools=string_pools)
    ra.chr_id = "chr1"
    ra.strand = "+"
    ra.corrected_exons = exons
    ra.corrected_introns = junctions_from_blocks(exons)
    ra.barcode = barcode
    ra.umi = umi
    ra.gene_assignment_type = assignment_type
    im = IsoformMatch(MatchClassification.full_splice_match, string_pools,
                      assigned_gene=gene_id, assigned_transcript=transcript_id)
    ra.isoform_matches = [im]
    ra.set_additional_attribute('transcript_type', 'protein_coding')
    ra.set_additional_attribute('polya_site', -1)
    ra.set_additional_attribute('cell_type', 'None')
    return ra


# -- file_naming tests --

class TestBarcode2BarcodeFileNaming:
    def test_prefix_format(self):
        result = umi_barcode2barcode_prefix("/out/sample.umi_done", 0)
        assert result == "/out/sample.umi_done.barcode_barcode_col0"

    def test_prefix_multiple_columns(self):
        assert umi_barcode2barcode_prefix("/x", 2) == "/x.barcode_barcode_col2"

    def test_lock_format(self):
        result = umi_barcode2barcode_global_lock("/out/sample.umi_done", 1)
        assert result == "/out/sample.umi_done.barcode_barcode_col1.lock"


# -- load_barcode2barcode_mapping tests --

class TestLoadBarcode2BarcodeMapping:
    def test_single_column(self):
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            f.write("AAAA\tspot_A\n")
            f.write("CCCC\tspot_A\n")
            f.write("GGGG\tspot_B\n")
            fname = f.name

        try:
            mapping = load_barcode2barcode_mapping(fname, barcode_col=0, spot_cols=[1])
            assert mapping["AAAA"] == ["spot_A"]
            assert mapping["CCCC"] == ["spot_A"]
            assert mapping["GGGG"] == ["spot_B"]
        finally:
            os.unlink(fname)

    def test_multi_column(self):
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            f.write("AAAA\tspot_A\tregion_1\n")
            f.write("CCCC\tspot_A\tregion_2\n")
            f.write("GGGG\tspot_B\tregion_1\n")
            fname = f.name

        try:
            mapping = load_barcode2barcode_mapping(fname, barcode_col=0, spot_cols=[1, 2])
            assert mapping["AAAA"] == ["spot_A", "region_1"]
            assert mapping["CCCC"] == ["spot_A", "region_2"]
            assert mapping["GGGG"] == ["spot_B", "region_1"]
        finally:
            os.unlink(fname)

    def test_custom_barcode_column(self):
        """Barcode column is not necessarily column 0."""
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            f.write("extra\tAAAA\tspot_X\n")
            f.write("extra\tCCCC\tspot_Y\n")
            fname = f.name

        try:
            mapping = load_barcode2barcode_mapping(fname, barcode_col=1, spot_cols=[2])
            assert mapping["AAAA"] == ["spot_X"]
            assert mapping["CCCC"] == ["spot_Y"]
        finally:
            os.unlink(fname)

    def test_empty_file(self):
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            fname = f.name

        try:
            mapping = load_barcode2barcode_mapping(fname, barcode_col=0, spot_cols=[1])
            assert mapping == {}
        finally:
            os.unlink(fname)

    def test_comments_and_blank_lines_skipped(self):
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            f.write("# header comment\n")
            f.write("\n")
            f.write("AAAA\tspot_A\n")
            f.write("# another comment\n")
            f.write("CCCC\tspot_B\n")
            fname = f.name

        try:
            mapping = load_barcode2barcode_mapping(fname, barcode_col=0, spot_cols=[1])
            assert len(mapping) == 2
            assert mapping["AAAA"] == ["spot_A"]
            assert mapping["CCCC"] == ["spot_B"]
        finally:
            os.unlink(fname)


# -- Grouper tests (barcode_barcode uses BarcodeSpotGrouper) --

class TestBarcode2BarcodeGrouper:
    """Test that barcode_barcode grouping works via BarcodeSpotGrouper."""

    def test_single_column_grouper(self):
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            f.write("AAAA\tspot_A\n")
            f.write("CCCC\tspot_A\n")
            f.write("GGGG\tspot_B\n")
            fname = f.name

        try:
            shared = SharedTableData(fname, read_id_column_index=0,
                                     group_id_column_indices=[1], delim='\t')
            grouper = BarcodeSpotGrouper(shared, column_index=0)

            alignment = MockAlignment()
            assert grouper.get_group_id(alignment, read_assignment=MockReadAssignment("AAAA")) == "spot_A"
            assert grouper.get_group_id(alignment, read_assignment=MockReadAssignment("CCCC")) == "spot_A"
            assert grouper.get_group_id(alignment, read_assignment=MockReadAssignment("GGGG")) == "spot_B"
        finally:
            os.unlink(fname)

    def test_multi_column_groupers(self):
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            f.write("AAAA\tspot_A\tregion_1\n")
            f.write("CCCC\tspot_B\tregion_1\n")
            fname = f.name

        try:
            shared = SharedTableData(fname, read_id_column_index=0,
                                     group_id_column_indices=[1, 2], delim='\t')
            g0 = BarcodeSpotGrouper(shared, column_index=0)
            g1 = BarcodeSpotGrouper(shared, column_index=1)

            alignment = MockAlignment()
            ra1 = MockReadAssignment("AAAA")
            ra2 = MockReadAssignment("CCCC")

            assert g0.get_group_id(alignment, read_assignment=ra1) == "spot_A"
            assert g0.get_group_id(alignment, read_assignment=ra2) == "spot_B"
            assert g1.get_group_id(alignment, read_assignment=ra1) == "region_1"
            assert g1.get_group_id(alignment, read_assignment=ra2) == "region_1"
        finally:
            os.unlink(fname)

    def test_unknown_barcode_returns_na(self):
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            f.write("AAAA\tspot_A\n")
            fname = f.name

        try:
            shared = SharedTableData(fname, read_id_column_index=0,
                                     group_id_column_indices=[1], delim='\t')
            grouper = BarcodeSpotGrouper(shared, column_index=0)
            alignment = MockAlignment()

            assert grouper.get_group_id(alignment, read_assignment=MockReadAssignment("ZZZZ")) == "NA"
        finally:
            os.unlink(fname)


# -- parse_grouping_spec tests for barcode_barcode --

class TestParseGroupingSpecBarcode2Barcode:
    """Test parse_grouping_spec with barcode_barcode spec."""

    class MockArgs:
        barcode2barcode = None

    class MockSample:
        barcodes_split_reads = "some_file"

    def test_single_column(self):
        from src.read_groups import parse_grouping_spec

        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            f.write("AAAA\tspot_A\n")
            f.write("CCCC\tspot_B\n")
            fname = f.name

        try:
            args = self.MockArgs()
            args.barcode2barcode = fname
            sample = self.MockSample()

            result = parse_grouping_spec("barcode_barcode", args, sample, "chr1")
            assert isinstance(result, BarcodeSpotGrouper)
        finally:
            os.unlink(fname)

    def test_multi_column_returns_list(self):
        from src.read_groups import parse_grouping_spec

        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            f.write("AAAA\tspot_A\tregion_1\n")
            f.write("CCCC\tspot_B\tregion_2\n")
            fname = f.name

        try:
            args = self.MockArgs()
            args.barcode2barcode = "%s:0:1,2" % fname
            sample = self.MockSample()

            result = parse_grouping_spec("barcode_barcode", args, sample, "chr1")
            assert isinstance(result, list)
            assert len(result) == 2
            assert all(isinstance(g, BarcodeSpotGrouper) for g in result)
        finally:
            os.unlink(fname)

    def test_missing_arg_exits(self):
        from src.read_groups import parse_grouping_spec

        args = self.MockArgs()
        args.barcode2barcode = None
        sample = self.MockSample()

        with pytest.raises(SystemExit):
            parse_grouping_spec("barcode_barcode", args, sample, "chr1")

    def test_missing_barcoded_reads_exits(self):
        from src.read_groups import parse_grouping_spec

        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            f.write("AAAA\tspot_A\n")
            fname = f.name

        try:
            args = self.MockArgs()
            args.barcode2barcode = fname
            sample = self.MockSample()
            sample.barcodes_split_reads = None

            with pytest.raises(SystemExit):
                parse_grouping_spec("barcode_barcode", args, sample, "chr1")
        finally:
            os.unlink(fname)


# -- get_grouping_strategy_names tests --

class TestGetGroupingStrategyNamesBarcode2Barcode:

    class MockArgs:
        barcode2barcode = None
        read_group = None

    def test_single_column_name(self):
        from src.read_groups import get_grouping_strategy_names

        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            f.write("AAAA\tspot_A\n")
            fname = f.name

        try:
            args = self.MockArgs()
            args.barcode2barcode = fname
            args.read_group = ["barcode_barcode"]

            names = get_grouping_strategy_names(args)
            assert names == ["barcode_barcode"]
        finally:
            os.unlink(fname)

    def test_multi_column_names(self):
        from src.read_groups import get_grouping_strategy_names

        args = self.MockArgs()
        args.barcode2barcode = "file.tsv:0:1,2,3"
        args.read_group = ["barcode_barcode"]

        names = get_grouping_strategy_names(args)
        assert names == ["barcode_barcode_col0", "barcode_barcode_col1", "barcode_barcode_col2"]


# -- get_grouping_pool_types tests --

class TestGetGroupingPoolTypesBarcode2Barcode:

    class MockArgs:
        barcode2barcode = None
        read_group = None

    def test_single_column_pool(self):
        from src.read_groups import get_grouping_pool_types

        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            f.write("AAAA\tspot_A\n")
            fname = f.name

        try:
            args = self.MockArgs()
            args.barcode2barcode = fname
            args.read_group = ["barcode_barcode"]

            pool_types = get_grouping_pool_types(args)
            assert pool_types == {0: 'barcode_barcode:0'}
        finally:
            os.unlink(fname)

    def test_multi_column_pools(self):
        from src.read_groups import get_grouping_pool_types

        args = self.MockArgs()
        args.barcode2barcode = "file.tsv:0:1,2"
        args.read_group = ["barcode_barcode"]

        pool_types = get_grouping_pool_types(args)
        assert pool_types == {0: 'barcode_barcode:0', 1: 'barcode_barcode:1'}


# -- UMIFilter barcode_remap tests --

class TestUMIFilterBarcodeRemap:
    string_pools = StringPoolManager()

    def test_init_stores_remap(self):
        remap = {"AAAA": "spot_X", "CCCC": "spot_Y"}
        f = UMIFilter(umi_length=10, edit_distance=3, barcode_remap=remap)
        assert f.barcode_remap is remap

    def test_init_default_remap_is_none(self):
        f = UMIFilter(umi_length=10, edit_distance=3)
        assert f.barcode_remap is None

    def test_remap_groups_by_spot(self):
        """Two barcodes mapping to the same spot should be grouped together for dedup."""
        remap = {"BC_A": "spot_1", "BC_B": "spot_1", "BC_C": "spot_2"}
        umi_filter = UMIFilter(umi_length=0, edit_distance=0, barcode_remap=remap)

        # Build gene_barcode_dict the same way process_single_chr does
        from collections import defaultdict
        gene_barcode_dict = defaultdict(lambda: defaultdict(list))

        ra1 = _make_read_assignment(self.string_pools, "r1", "BC_A", "UMI1", "gene1", "tx1",
                                    [(100, 200), (300, 400)])
        ra2 = _make_read_assignment(self.string_pools, "r2", "BC_B", "UMI1", "gene1", "tx1",
                                    [(100, 200), (300, 400)])
        ra3 = _make_read_assignment(self.string_pools, "r3", "BC_C", "UMI2", "gene1", "tx1",
                                    [(100, 200), (300, 400)])

        for ra in [ra1, ra2, ra3]:
            barcode = ra.barcode
            gene_id = ra.isoform_matches[0].assigned_gene
            grouping_key = barcode
            if umi_filter.barcode_remap and barcode in umi_filter.barcode_remap:
                grouping_key = umi_filter.barcode_remap[barcode]
            gene_barcode_dict[gene_id][grouping_key].append(ra)

        # BC_A and BC_B both map to spot_1, so they should be in the same group
        assert len(gene_barcode_dict["gene1"]["spot_1"]) == 2
        assert len(gene_barcode_dict["gene1"]["spot_2"]) == 1

    def test_no_remap_groups_by_barcode(self):
        """Without remap, reads group by original barcode."""
        umi_filter = UMIFilter(umi_length=0, edit_distance=0, barcode_remap=None)

        from collections import defaultdict
        gene_barcode_dict = defaultdict(lambda: defaultdict(list))

        ra1 = _make_read_assignment(self.string_pools, "r1", "BC_A", "UMI1", "gene1", "tx1",
                                    [(100, 200), (300, 400)])
        ra2 = _make_read_assignment(self.string_pools, "r2", "BC_B", "UMI1", "gene1", "tx1",
                                    [(100, 200), (300, 400)])

        for ra in [ra1, ra2]:
            barcode = ra.barcode
            gene_id = ra.isoform_matches[0].assigned_gene
            grouping_key = barcode
            if umi_filter.barcode_remap and barcode in umi_filter.barcode_remap:
                grouping_key = umi_filter.barcode_remap[barcode]
            gene_barcode_dict[gene_id][grouping_key].append(ra)

        # Without remap, each barcode stays separate
        assert len(gene_barcode_dict["gene1"]) == 2
        assert len(gene_barcode_dict["gene1"]["BC_A"]) == 1
        assert len(gene_barcode_dict["gene1"]["BC_B"]) == 1

    def test_remap_unknown_barcode_stays_as_is(self):
        """Barcodes not in the remap dict should keep their original barcode as key."""
        remap = {"BC_A": "spot_1"}
        umi_filter = UMIFilter(umi_length=0, edit_distance=0, barcode_remap=remap)

        from collections import defaultdict
        gene_barcode_dict = defaultdict(lambda: defaultdict(list))

        ra1 = _make_read_assignment(self.string_pools, "r1", "BC_A", "UMI1", "gene1", "tx1",
                                    [(100, 200), (300, 400)])
        ra2 = _make_read_assignment(self.string_pools, "r2", "BC_UNKNOWN", "UMI2", "gene1", "tx1",
                                    [(100, 200), (300, 400)])

        for ra in [ra1, ra2]:
            barcode = ra.barcode
            gene_id = ra.isoform_matches[0].assigned_gene
            grouping_key = barcode
            if umi_filter.barcode_remap and barcode in umi_filter.barcode_remap:
                grouping_key = umi_filter.barcode_remap[barcode]
            gene_barcode_dict[gene_id][grouping_key].append(ra)

        assert "spot_1" in gene_barcode_dict["gene1"]
        assert "BC_UNKNOWN" in gene_barcode_dict["gene1"]

    def test_remap_dedup_same_umi_different_barcodes(self):
        """Reads with same UMI, different barcodes but same spot should be deduped together."""
        remap = {"BC_A": "spot_1", "BC_B": "spot_1"}
        umi_filter = UMIFilter(umi_length=0, edit_distance=0, barcode_remap=remap)

        ra1 = _make_read_assignment(self.string_pools, "r1", "BC_A", "AAAA", "gene1", "tx1",
                                    [(100, 200), (300, 400)])
        ra2 = _make_read_assignment(self.string_pools, "r2", "BC_B", "AAAA", "gene1", "tx1",
                                    [(100, 200), (300, 400)])

        # Group by spot via remap
        from collections import defaultdict
        gene_barcode_dict = defaultdict(lambda: defaultdict(list))
        for ra in [ra1, ra2]:
            barcode = ra.barcode
            gene_id = ra.isoform_matches[0].assigned_gene
            grouping_key = remap.get(barcode, barcode)
            gene_barcode_dict[gene_id][grouping_key].append(ra)

        # Both reads are in spot_1 with same UMI — _process_duplicates should dedup
        result = umi_filter._process_duplicates(gene_barcode_dict["gene1"]["spot_1"])
        assert len(result) == 1  # Deduped to 1 representative read


# -- CLI argument tests --

class TestBarcode2BarcodeArgParsing:
    """Test that --barcode2barcode argument is parsed correctly."""

    def test_arg_added(self):
        """Verify the argument exists in the parser."""
        import sys
        sys.argv = ['isoquant.py', '--full_help']
        # We just test that parse_barcode2spot_spec works on barcode2barcode-style specs
        filename, bc_col, spot_cols = parse_barcode2spot_spec("mapping.tsv:0:1,2")
        assert filename == "mapping.tsv"
        assert bc_col == 0
        assert spot_cols == [1, 2]


if __name__ == '__main__':
    pytest.main([__file__, '-v'])



############################################################################
# Copyright (c) 2024-2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################
import threading

from .common import TranscriptNaming


class SimpleIDDistributor(object):
    def __init__(self):
        self.value = 0

    def increment(self):
        self.value += 1
        return self.value


class ExcludingIdDistributor(SimpleIDDistributor):
    def __init__(self, genedb, chr_id):
        SimpleIDDistributor.__init__(self)
        self.forbidden_ids = set()
        if not genedb:
            return

        for g in genedb.region(seqid=chr_id, start=1, featuretype="gene"):
            if g.id.startswith(TranscriptNaming.novel_gene_prefix):
                try:
                    gene_num = int(g.id.split("_")[-1])
                    self.forbidden_ids.add(gene_num)
                except IndexError:
                    pass
                except ValueError:
                    pass

        transcript_num_start_pos = len(TranscriptNaming.transcript_prefix)
        for t in genedb.region(seqid=chr_id, start=1, featuretype=("transcript", "mRNA")):
            if t.id.startswith(TranscriptNaming.transcript_prefix):
                try:
                    transcript_num = int(t.id.split(".")[0][transcript_num_start_pos:])
                    self.forbidden_ids.add(transcript_num)
                except IndexError:
                    pass
                except ValueError:
                    pass

    def increment(self):
        self.value += 1
        while self.value in self.forbidden_ids:
            self.value += 1
        return self.value


class FeatureIdStorage:
    def __init__(self, id_distributor, genedb = None, chr_id = None, feature = "exon"):
        self.id_distributor = id_distributor
        self.id_dict = {}
        self.feature_name = feature
        if not genedb or not chr_id:
            return

        id_attribute = feature + "_id"
        for f in genedb.region(seqid=chr_id, start=1, featuretype=feature):
            if id_attribute in f.attributes:
                feature_tuple = (chr_id, f.start, f.end, f.strand)
                try:
                    self.id_dict[feature_tuple] = f.attributes[id_attribute][0]
                except IndexError:
                    pass

    def get_id(self, chr_id, feature, strand):
        feature_tuple = (chr_id, feature[0], feature[1], strand)
        if feature_tuple not in self.id_dict:
            feature_id = self.id_distributor.increment()
            self.id_dict[feature_tuple] = chr_id + ".%d" % feature_id
        else:
            feature_id =  self.id_dict[feature_tuple]

        return feature_id


class AtomicIDDistributor(object):
    def __init__(self):
        self.value = 0
        self._lock = threading.Lock()

    def increment(self):
        with self._lock:
            self.value += 1
            return self.value



############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2020-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging
import os
from collections import defaultdict
from collections import namedtuple
import gzip

from .common import max_range
from .gene_info import TranscriptModel, GeneInfo

logger = logging.getLogger('IsoQuant')


def validate_exons(novel_exons):
    return novel_exons == sorted(novel_exons) and all(0 < x[0] <= x[1] for x in novel_exons)


class VoidTranscriptPrinter:
    def dump(self, transcript_model_constructor, transcript_model_storage=None):
        pass


class GFFPrinter:
    exon_id_dict = {}

    def __init__(self, outf_prefix, sample_name, exon_id_storage,
                 gtf_suffix = ".transcript_models.gtf",
                 r2t_suffix = ".transcript_model_reads.tsv",
                 output_r2t = True,
                 check_canonical = False,
                 header = "",
                 gzipped = False):
        self.model_fname = os.path.join(outf_prefix, sample_name + gtf_suffix)
        self.exon_id_storage = exon_id_storage
        self.printed_gene_ids = set()
        self.out_gff = open(self.model_fname, "w")
        if header:
            self.out_gff.write("# " + sample_name + " IsoQuant generated GTF\n" + header)
            self.out_gff.flush()

        self.output_r2t = output_r2t
        if self.output_r2t:
            self.r2t_fname = os.path.join(outf_prefix, sample_name + r2t_suffix)
            if gzipped:
                self.out_r2t = gzip.open(self.r2t_fname + ".gz", "wt")
            else:
                self.out_r2t = open(self.r2t_fname, "w")
            if header:
                self.out_r2t.write("read_id\ttranscript_id\n")

        self.check_canonical = check_canonical

    def __del__(self):
        self.out_gff.close()
        if self.output_r2t:
            self.out_r2t.close()

    def dump(self, gene_info, transcript_model_storage):
        if not transcript_model_storage:
            return

        gene_to_model_dict = defaultdict(list)
        gene_regions = {}
        if not gene_info.empty():
            gene_regions = gene_info.get_gene_regions()
        GFFGeneInfo = namedtuple("GFFGeneInfo", ("chr_id", "strand", "gene_region"))
        gene_info_dict = {}

        for i, model in enumerate(transcript_model_storage):
            if not validate_exons(model.exon_blocks):
                logger.warning("Transcript model %s has incorrect coordinates and will be ignored: %s" %
                               (model.transcript_id, str(model.exon_blocks)))
                continue
            gene_id = model.gene_id
            gene_to_model_dict[gene_id].append(i)

            transcript_region = (model.exon_blocks[0][0], model.exon_blocks[-1][1])
            if gene_id not in gene_info_dict:
                assert model.chr_id == gene_info.chr_id
                gene_range = max_range(gene_regions[gene_id], transcript_region) if gene_id in gene_regions else transcript_region
                gene_info_dict[gene_id] = GFFGeneInfo(model.chr_id, model.strand, gene_range)
            else:
                gene_record = gene_info_dict[gene_id]
                assert model.chr_id == gene_record.chr_id
                if model.strand != gene_record.strand:
                    logger.warning("Gene and transcript records have unequal strands: %s: %s, %s: %s" %
                                   (gene_id, gene_record.strand, model.transcript_id, model.strand))
                gene_info_dict[gene_id] = GFFGeneInfo(model.chr_id, model.strand,
                                                      max_range(gene_record.gene_region, transcript_region))

        gene_order = sorted([(g, gene_info_dict[g].gene_region) for g in gene_info_dict.keys()], key=lambda x:x[1])

        for gene_id, coords in gene_order:
            if gene_id not in self.printed_gene_ids:
                gene_additiional_info = ""
                if gene_info and gene_id in gene_info.feature_attributes:
                    gene_additiional_info = gene_info.feature_attributes[gene_id]
                source = "IsoQuant"
                if gene_info and gene_id in gene_info.sources:
                    source = gene_info.sources[gene_id]
                gene_line = '%s\t%s\tgene\t%d\t%d\t.\t%s\t.\tgene_id "%s"; transcripts "%d"; %s\n' % \
                        (gene_info_dict[gene_id].chr_id, source, coords[0], coords[1], gene_info_dict[gene_id].strand,
                         gene_id, len(gene_to_model_dict[gene_id]), gene_additiional_info)
                self.out_gff.write(gene_line)
                self.printed_gene_ids.add(gene_id)

            for model_index in gene_to_model_dict[gene_id]:
                model = transcript_model_storage[model_index]
                assert model.gene_id == gene_id

                if not model.check_additional("exons"):
                    model.add_additional_attribute("exons", str(len(model.exon_blocks)))
                transcript_additiional_info = ""
                if gene_info and model.transcript_id in gene_info.feature_attributes:
                    transcript_additiional_info = " " + gene_info.feature_attributes[model.transcript_id]

                transcript_line = '%s\t%s\ttranscript\t%d\t%d\t.\t%s\t.\tgene_id "%s"; transcript_id "%s"; %s\n' \
                                  % (model.chr_id,  model.source, model.exon_blocks[0][0], model.exon_blocks[-1][1],
                                     model.strand, model.gene_id, model.transcript_id,
                                     model.additional_attributes_str() + transcript_additiional_info)
                self.out_gff.write(transcript_line)

                prefix_columns = "%s\t%s\t" % (model.chr_id, model.source)
                suffix_columns = '.\t%s\t.\tgene_id "%s"; transcript_id "%s";' % \
                                 (model.strand, model.gene_id, model.transcript_id)

                exons_to_print = []
                exons_to_print += model.other_features
                for e in model.exon_blocks:
                    exons_to_print.append((e[0], e[1], 'exon'))
                exons_to_print = sorted(exons_to_print, reverse=True) if model.strand == '-' else sorted(exons_to_print)
                for i, e in enumerate(exons_to_print):
                    exon_str_id = self.exon_id_storage.get_id(model.chr_id, e, model.strand)

                    exon_id = model.transcript_id + "_%d_%d_%s" % (e[0], e[1], model.strand)
                    exon_additiional_info = ""
                    if gene_info and exon_id in gene_info.feature_attributes:
                        exon_additiional_info = " " + gene_info.feature_attributes[model.transcript_id]
                    feature_type = e[2]
                    self.out_gff.write(prefix_columns + "%s\t%d\t%d\t" % (feature_type, e[0], e[1]) + suffix_columns +
                                       ' exon_number "%d"; exon_id "%s"; %s\n' % ((i + 1), exon_str_id, exon_additiional_info))
        self.out_gff.flush()

    def dump_read_assignments(self, transcript_model_constructor):
        # write read_id -> transcript_id map
        if not self.output_r2t:
            return
        for model_id in transcript_model_constructor.transcript_read_ids.keys():
            read_assignments = transcript_model_constructor.transcript_read_ids[model_id]
            for a in read_assignments:
                self.out_r2t.write("%s\t%s\n" % (a.read_id, model_id))
        for read_id in transcript_model_constructor.read_assignment_counts.keys():
            if transcript_model_constructor.read_assignment_counts[read_id] == 0:
                self.out_r2t.write("%s\t%s\n" % (read_id, "*"))


def create_extended_storage(genedb, chr_id, chr_record, novel_model_storage):
    all_models = []
    gene_list = list(genedb.region(seqid=chr_id, start=1, featuretype="gene"))
    if not gene_list:
        for m in novel_model_storage:
            all_models.append(m)
        return all_models, GeneInfo.from_region(chr_id, 1, len(chr_record), chr_record=chr_record)
    gene_info = GeneInfo(gene_list, genedb, prepare_profiles=False)
    gene_info.set_reference_sequence(1, len(chr_record), chr_record)
    for isoform_id in gene_info.all_isoforms_exons.keys():
        all_models.append(TranscriptModel.from_reference_transcript(gene_info, isoform_id))
    for m in novel_model_storage:
        all_models.append(m)

    return all_models, gene_info



############################################################################
# Copyright (c) 2023-2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

"""
Shared memory k-mer indexer for parallel processing of large barcode sets.

Uses multiprocessing.shared_memory to share index across worker processes.
Optimized with Numba JIT compilation when available.
"""

import gc
import logging
import math
import numpy
from collections import defaultdict
from multiprocessing import shared_memory
from typing import List, Tuple
from ..common import bit_to_str, str_to_2bit

logger = logging.getLogger('IsoQuant')

# Try to import numba for JIT compilation
try:
    from numba import njit
    NUMBA_AVAILABLE = True
except (ImportError, SystemError):
    # SystemError can occur with broken numba installations
    NUMBA_AVAILABLE = False


def _count_kmers_python(seqs, shifts, mask, kmer_counts, num_kmers_per_seq):
    """Pure Python fallback for k-mer counting."""
    for seq_idx in range(len(seqs)):
        seq = int(seqs[seq_idx])
        for k_pos in range(num_kmers_per_seq):
            kmer_idx = (seq >> int(shifts[k_pos])) & mask
            kmer_counts[kmer_idx] += 1


def _populate_index_python(seqs, shifts, mask, index_ranges, index, num_kmers_per_seq):
    """Pure Python fallback for index population."""
    total_kmers = len(index_ranges) - 1
    write_positions = numpy.zeros(total_kmers, dtype=numpy.uint64)

    for seq_idx in range(len(seqs)):
        seq = int(seqs[seq_idx])
        for k_pos in range(num_kmers_per_seq):
            kmer_idx = (seq >> int(shifts[k_pos])) & mask
            write_pos = int(index_ranges[kmer_idx] + write_positions[kmer_idx])
            index[write_pos] = seq_idx
            write_positions[kmer_idx] += 1


if NUMBA_AVAILABLE:
    @njit(cache=True)
    def _count_kmers_numba(seqs, shifts, mask, kmer_counts, num_kmers_per_seq):
        """Numba JIT-compiled k-mer counting."""
        for seq_idx in range(len(seqs)):
            seq = seqs[seq_idx]
            for k_pos in range(num_kmers_per_seq):
                kmer_idx = (seq >> shifts[k_pos]) & mask
                kmer_counts[kmer_idx] += 1

    @njit(cache=True)
    def _populate_index_numba(seqs, shifts, mask, index_ranges, index, num_kmers_per_seq):
        """Numba JIT-compiled index population - O(n) direct writes."""
        total_kmers = len(index_ranges) - 1
        write_positions = numpy.zeros(total_kmers, dtype=numpy.uint64)

        for seq_idx in range(len(seqs)):
            seq = seqs[seq_idx]
            for k_pos in range(num_kmers_per_seq):
                kmer_idx = (seq >> shifts[k_pos]) & mask
                write_pos = index_ranges[kmer_idx] + write_positions[kmer_idx]
                index[write_pos] = seq_idx
                write_positions[kmer_idx] += 1

    _count_kmers = _count_kmers_numba
    _populate_index = _populate_index_numba
else:
    _count_kmers = _count_kmers_python
    _populate_index = _populate_index_python


class SharedMemoryIndexInfo:
    """
    Metadata container for passing shared memory index info between processes.

    Stores shared memory block names and index parameters for reconstruction
    in worker processes.
    """

    def __init__(self, barcode_count: int, kmer_size: int, seq_len: int,
                 index_size: int, barcodes_sm_name: str, index_sm_name: str,
                 index_range_sm_name: str):
        self.barcode_count = barcode_count
        self.kmer_size = kmer_size
        self.seq_len = seq_len
        self.index_size = index_size
        self.barcodes_sm_name = barcodes_sm_name
        self.index_sm_name = index_sm_name
        self.index_range_sm_name = index_range_sm_name

    def __getstate__(self):
        return (self.barcode_count, self.kmer_size, self.seq_len,
                self.index_size, self.barcodes_sm_name,
                self.index_sm_name, self.index_range_sm_name)

    def __setstate__(self, state):
        self.barcode_count = state[0]
        self.kmer_size = state[1]
        self.seq_len = state[2]
        self.index_size = state[3]
        self.barcodes_sm_name = state[4]
        self.index_sm_name = state[5]
        self.index_range_sm_name = state[6]


class SharedMemoryArray2BitKmerIndexer:
    """
    Shared memory k-mer indexer for parallel processing of large barcode sets.

    Allocates index data in shared memory so worker processes can access
    the same index without copying. Uses Numba JIT compilation when available
    for fast index construction.

    Best for barcode sets > 1M where memory sharing is critical.
    """

    SEQ_DTYPE = numpy.uint64
    KMER_DTYPE = numpy.uint64
    INDEX_DTYPE = numpy.uint64

    def __init__(self, known_bin_seq: list, kmer_size: int = 12, seq_len: int = 25):
        """
        Initialize shared memory k-mer index.

        Args:
            known_bin_seq: Pre-encoded sequences as integers (use str_to_2bit)
            kmer_size: Length of k-mers
            seq_len: Length of sequences (all must be same length)
        """
        self.main_instance = True
        self.k = kmer_size
        total_kmers = int(math.pow(4, self.k))
        self.mask = (1 << (2 * self.k)) - 1
        self.seq_len = seq_len
        self.seq_mask = (1 << (2 * self.seq_len)) - 1
        self.total_sequences = len(known_bin_seq)

        # Handle empty case
        if self.total_sequences == 0:
            self.index_size = 0
            self.barcodes_shared_memory = None
            self.index_shared_memory = None
            self.index_ranges_shared_memory = None
            self.known_bin_seq = numpy.array([], dtype=self.SEQ_DTYPE)
            self.index = numpy.array([], dtype=self.KMER_DTYPE)
            self.index_ranges = numpy.zeros(total_kmers + 1, dtype=self.INDEX_DTYPE)
            return

        # Convert to numpy array for vectorized operations
        seqs = numpy.array(known_bin_seq, dtype=self.SEQ_DTYPE)

        # Precompute bit shifts for k-mer extraction
        num_kmers_per_seq = seq_len - kmer_size + 1
        shifts = numpy.array([(seq_len - kmer_size - i) * 2 for i in range(num_kmers_per_seq)],
                             dtype=numpy.uint64)

        logger.info("Building k-mer index for %d sequences (Numba: %s)",
                    self.total_sequences, "enabled" if NUMBA_AVAILABLE else "disabled")

        # Pass 1: Count k-mer occurrences (Numba JIT if available)
        kmer_counts = numpy.zeros(total_kmers, dtype=numpy.uint64)
        _count_kmers(seqs, shifts, self.mask, kmer_counts, num_kmers_per_seq)
        logger.info("K-mer counting complete")

        # Compute index size from counts
        self.index_size = int(numpy.sum(kmer_counts))

        # Allocate barcodes shared memory
        self.barcodes_shared_memory = shared_memory.SharedMemory(
            create=True, size=self.total_sequences * self.SEQ_DTYPE().nbytes)
        self.known_bin_seq = numpy.ndarray(
            shape=(self.total_sequences,), dtype=self.SEQ_DTYPE,
            buffer=self.barcodes_shared_memory.buf)
        self.known_bin_seq[:] = seqs[:]

        # Allocate index shared memory
        self.index_shared_memory = shared_memory.SharedMemory(
            create=True, size=self.index_size * self.KMER_DTYPE().nbytes)
        self.index = numpy.ndarray(
            shape=(self.index_size,), dtype=self.KMER_DTYPE,
            buffer=self.index_shared_memory.buf)

        # Allocate index_ranges shared memory
        self.index_ranges_shared_memory = shared_memory.SharedMemory(
            create=True, size=(total_kmers + 1) * self.INDEX_DTYPE().nbytes)
        self.index_ranges = numpy.ndarray(
            shape=(total_kmers + 1,), dtype=self.INDEX_DTYPE,
            buffer=self.index_ranges_shared_memory.buf)

        # Compute ranges from counts using cumsum
        self.index_ranges[0] = 0
        self.index_ranges[1:] = numpy.cumsum(kmer_counts)

        # Pass 2: Populate index using direct writes (Numba JIT if available)
        _populate_index(seqs, shifts, self.mask, self.index_ranges, self.index, num_kmers_per_seq)
        logger.info("K-mer index population complete")

        # Explicitly delete temporary arrays to free memory before forking
        del seqs, kmer_counts, shifts
        gc.collect()

    def __del__(self):
        barcodes_sm = getattr(self, 'barcodes_shared_memory', None)
        if barcodes_sm is not None:
            barcodes_sm.close()
            if getattr(self, 'main_instance', False):
                barcodes_sm.unlink()
        index_sm = getattr(self, 'index_shared_memory', None)
        if index_sm is not None:
            index_sm.close()
            if getattr(self, 'main_instance', False):
                index_sm.unlink()
        ranges_sm = getattr(self, 'index_ranges_shared_memory', None)
        if ranges_sm is not None:
            ranges_sm.close()
            if getattr(self, 'main_instance', False):
                ranges_sm.unlink()

    @classmethod
    def from_sharable_info(cls, shared_mem_index_info: SharedMemoryIndexInfo) -> 'SharedMemoryArray2BitKmerIndexer':
        """Reconstruct indexer in worker process from shared memory info."""
        kmer_index = cls.__new__(cls)
        kmer_index.main_instance = False
        kmer_index.k = shared_mem_index_info.kmer_size
        kmer_index.mask = (1 << (2 * kmer_index.k)) - 1
        kmer_index.seq_len = shared_mem_index_info.seq_len
        kmer_index.seq_mask = (1 << (2 * kmer_index.seq_len)) - 1
        kmer_index.index_size = shared_mem_index_info.index_size
        kmer_index.total_sequences = shared_mem_index_info.barcode_count
        total_kmers = int(math.pow(4, kmer_index.k))

        kmer_index.barcodes_shared_memory = shared_memory.SharedMemory(
            create=False, name=shared_mem_index_info.barcodes_sm_name)
        kmer_index.known_bin_seq = numpy.ndarray(
            shape=(kmer_index.total_sequences,),
            dtype=SharedMemoryArray2BitKmerIndexer.SEQ_DTYPE,
            buffer=kmer_index.barcodes_shared_memory.buf)

        kmer_index.index_shared_memory = shared_memory.SharedMemory(
            create=False, name=shared_mem_index_info.index_sm_name)
        kmer_index.index = numpy.ndarray(
            shape=(kmer_index.index_size,),
            dtype=kmer_index.KMER_DTYPE,
            buffer=kmer_index.index_shared_memory.buf)

        kmer_index.index_ranges_shared_memory = shared_memory.SharedMemory(
            create=False, name=shared_mem_index_info.index_range_sm_name)
        kmer_index.index_ranges = numpy.ndarray(
            shape=(total_kmers + 1,),
            dtype=kmer_index.INDEX_DTYPE,
            buffer=kmer_index.index_ranges_shared_memory.buf)

        return kmer_index

    def get_sharable_info(self) -> SharedMemoryIndexInfo:
        """Get info needed to reconstruct indexer in worker process."""
        return SharedMemoryIndexInfo(
            self.total_sequences, self.k, self.seq_len, self.index_size,
            self.barcodes_shared_memory.name, self.index_shared_memory.name,
            self.index_ranges_shared_memory.name)

    def _get_kmer_bin_indexes(self, bin_seq: int):
        """Extract k-mer indices from a 2-bit encoded sequence."""
        for i in range(self.seq_len - self.k + 1):
            yield (bin_seq >> ((self.seq_len - self.k - i) * 2)) & self.mask

    def empty(self) -> bool:
        """Check if index is empty."""
        return len(self.known_bin_seq) == 0

    def get_occurrences(self, sequence: str, max_hits: int = 0, min_kmers: int = 1,
                       hits_delta: int = 1, ignore_equal: bool = False) -> List[Tuple[str, int, List[int]]]:
        """
        Find indexed sequences with shared k-mers.

        Args:
            sequence: Query sequence (string, will be converted to 2-bit)
            max_hits: Maximum number of results (0 = unlimited)
            min_kmers: Minimum shared k-mers required
            hits_delta: Include results within this many k-mers of top hit
            ignore_equal: Skip exact matches

        Returns:
            List of (sequence_str, shared_kmer_count, kmer_positions) tuples
        """
        barcode_counts = defaultdict(int)
        barcode_positions = defaultdict(list)

        seq = str_to_2bit(sequence)
        for pos, kmer_idx in enumerate(self._get_kmer_bin_indexes(seq)):
            start_index = self.index_ranges[kmer_idx]
            end_index = self.index_ranges[kmer_idx + 1]
            for barcode_index in range(start_index, end_index):
                barcode = int(self.known_bin_seq[self.index[barcode_index]])
                barcode_counts[barcode] += 1
                barcode_positions[barcode].append(pos)

        result = []
        for barcode in barcode_counts.keys():
            count = barcode_counts[barcode]
            if count < min_kmers:
                continue
            if ignore_equal and barcode == seq:
                continue
            result.append((barcode, count, barcode_positions[barcode]))

        if not result:
            return []

        top_hits = max(result, key=lambda x: x[1])[1]
        result = filter(lambda x: x[1] >= top_hits - hits_delta, result)
        result = sorted(result, reverse=True, key=lambda x: x[1])

        if max_hits == 0:
            return [(bit_to_str(x[0], self.seq_len), x[1], x[2]) for x in result]
        return [(bit_to_str(x[0], self.seq_len), x[1], x[2]) for x in list(result)[:max_hits]]




############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

from collections import namedtuple

import pytest
import gffutils
import os
from functools import partial

from src.common import equal_ranges, overlaps_at_least
from src.isoform_assignment import MatchEventSubtype, ReadAssignmentType
from src.long_read_assigner import (
    AmbiguityResolvingMethod,
    IsoformDiff,
    LongReadAssigner,
)
from src.long_read_profiles import (
    CombinedReadProfiles,
    OverlappingFeaturesProfileConstructor,
    MappedReadProfile,
    NonOverlappingFeaturesProfileConstructor,
)
from src.gene_info import GeneInfo
from src.polya_finder import PolyAInfo
from src.string_pools import StringPoolManager


class Params:
    def __init__(self, delta):
        self.delta = delta
        self.minor_exon_extension = 15
        self.major_exon_extension = 100
        self.min_abs_exon_overlap = 10
        self.min_rel_exon_overlap = 0.2
        self.max_suspicious_intron_abs_len = 0
        self.max_suspicious_intron_rel_len = 0
        self.max_fake_terminal_exon_len = 0
        self.micro_intron_length = 0
        self.max_intron_abs_diff = 5
        self.max_intron_rel_diff = 0.1
        self.apa_delta = 50
        self.minimal_exon_overlap = 5
        self.minimal_intron_absence_overlap = 5
        self.max_intron_shift = 10
        self.max_missed_exon_len = 10
        self.resolve_ambiguous = AmbiguityResolvingMethod.monoexon_only
        self.correct_minor_errors = True

IntronProfiles = namedtuple("IntronProfiles", ("features", ))
GeneInfoTuple = namedtuple("GeneInfo", ("intron_profiles", "start", "end"))


class TestMatchProfileAndFindMatchingIsoforms:
    gene_info = GeneInfoTuple(IntronProfiles([(50, 60), (80, 100), (80, 110), (200, 210)]), 0, 300)
    string_pools = StringPoolManager()
    assigner = LongReadAssigner(gene_info, Params(0), string_pools)  # we need no params here

    @pytest.mark.parametrize("read_gene_profile, isoform_profiles, hint, expected",
                             [([], dict(id1=[1], id2=[-1]), None, [IsoformDiff("id2", 0), IsoformDiff("id1", 0)])])
    def test_empty(self, read_gene_profile, isoform_profiles, hint, expected):
        with pytest.raises(AssertionError):
            self.check(read_gene_profile, isoform_profiles, hint, expected)

    @pytest.mark.parametrize("read_gene_profile, isoform_profiles, hint, expected",
                             [([-1, 1, -1, 0], dict(id1=[-1, 1, 1, -1], id2=[-1, 1 ,-2], id3=[-1, 1, 1, 1]),
                               {"id2", "id3"}, [IsoformDiff("id2", -1), IsoformDiff("id3", 1)])])
    def test_different_length(self, read_gene_profile, isoform_profiles, hint, expected):
        with pytest.raises(AssertionError):
            self.check(read_gene_profile, isoform_profiles, hint, expected)

    def check(self, read_gene_profile, isoform_profiles, hint, expected):
        assert self.assigner.match_profile(read_gene_profile, isoform_profiles, hint) == expected
        expected = sorted([x[0] for x in expected if x[1] == 0])
        assert self.assigner.find_matching_isoforms(MappedReadProfile(read_gene_profile, None, None, None),
                                                                isoform_profiles, hint) == expected

    @pytest.mark.parametrize("read_gene_profile, isoform_profiles, hint, expected",
                             [([-1, 1, -1, 0], dict(id1=[-1, 1, -1, -1], id2=[-1, 1, -1, -2]),
                               None, [IsoformDiff("id1", 0), IsoformDiff("id2", 0)]),
                              ([0, 1, -1, 1], dict(id1=[-1, 1, -1, 1], id2=[-2, 1, -1, 1]),
                               None, [IsoformDiff("id1", 0), IsoformDiff("id2", 0)])])
    def test_all_equals(self, read_gene_profile, isoform_profiles, hint, expected):
        self.check(read_gene_profile, isoform_profiles, hint, expected)

    @pytest.mark.parametrize("read_gene_profile, isoform_profiles, hint, expected",
                             [([-1, 1, -1, 0], dict(id1=[-1, 1, -1, -1], id2=[1, 1, 1, -1]),
                               None, [IsoformDiff("id1", 0), IsoformDiff("id2", 2)]),
                              ([0, 1, 1, -1, 0], dict(id1=[-1, 1, 1,  1, -2], id2=[-2, 1, 1, -1, -2],
                                                       id3=[ 1, -1, 1, -2, -2]),
                               None, [IsoformDiff("id2", 0), IsoformDiff("id1", 1), IsoformDiff("id3", 2)])])
    def test_some_equals(self, read_gene_profile, isoform_profiles, hint, expected):
        self.check(read_gene_profile, isoform_profiles, hint, expected)

    @pytest.mark.parametrize("read_gene_profile, isoform_profiles, hint, expected",
                             [([-1, 1, -1, 0], dict(id1=[-2, 1, -1, -1], id2=[1, 1, 1, 1]),
                               None, [IsoformDiff("id1", 1), IsoformDiff("id2", 2)]),
                              ([-1, 1, 1, -1, 0], dict(id1=[-1, 1, 1, 1, 1], id2=[1, 1, -1, -1, -1],
                                                       id3=[-2, -1, -1, 1, -2]),
                               None, [IsoformDiff("id1", 1), IsoformDiff("id2", 2), IsoformDiff("id3", 4)])])
    def test_no_equals(self, read_gene_profile, isoform_profiles, hint, expected):
        self.check(read_gene_profile, isoform_profiles, hint, expected)

    @pytest.mark.parametrize("read_gene_profile, isoform_profiles, hint, expected",
                             [([-1, 1, 1, 0], dict(id1=[-1, 1, 1, -1], id2=[1, 1, 1, -1], id3=[-1, -1, -1, 1]),
                               {"id2", "id3"}, [IsoformDiff("id2", 1), IsoformDiff("id3", 2)]),  # skip matched return another one
                              ([-1, 1, -1, 0], dict(id1=[-1, 1, 1, -2], id2=[-1, -1, 1, -2], id3=[-1, 1, -1, -2]),
                               {"id3"}, [IsoformDiff("id3", 0)])])  # matched
    def test_hint(self, read_gene_profile, isoform_profiles, hint, expected):
        self.check(read_gene_profile, isoform_profiles, hint, expected)


class TestCompareJunctions:
    gene_info = GeneInfoTuple(IntronProfiles([(50, 60), (80, 100), (80, 110), (80, 150), (200, 210)]), 0, 300)
    string_pools = StringPoolManager()

    @pytest.mark.parametrize("junctions, region, delta, expected",
                             [([(50, 60), (80, 100), (200, 210)], (1, 1), 3, [0, 1, -1, 0]),
                              ([(50, 60), (80, 100), (200, 210)], (0, 1), 3, [1, 1, -1, 0])])
    def test_profile_for_junctions_introns(self, junctions, region, delta,  expected):
        assigner = LongReadAssigner(self.gene_info, Params(delta), self.string_pools)
        profile = assigner.intron_comparator.profile_for_junctions_introns(junctions, region)
        assert profile.gene_profile == expected

    @pytest.mark.parametrize("junctions, region, delta, expected",
                             [([(50, 60), (82, 100), (200, 210)], (1, 1), 3, True),
                              ([(48, 61), (80, 98), (200, 210)], (0, 1), 3, True),
                              ([(48, 61), (80, 98), (160, 220)], (0, 1), 3, True),
                              ([(48, 66), (80, 99), (200, 210)], (0, 1), 3, False)])
    def test_profile_for_junctions_introns(self, junctions, region, delta, expected):
        assigner = LongReadAssigner(self.gene_info, Params(delta), self.string_pools)
        assert assigner.intron_comparator.are_known_introns(junctions, region) == expected

    @pytest.mark.parametrize("read_intron_read_profile, read_region, read_introns, delta, expected",
                             [([0, 1, 1], (0, 300), [(30, 60), (82, 100), (200, 210)], 3, MatchEventSubtype.extra_intron_flanking_left),
                              ([0, 1, 1], (0, 300), [(5, 60), (82, 100), (200, 210)], 3, MatchEventSubtype.fake_terminal_exon_left),
                              ([1, 1, 0], (0, 300), [(50, 60), (82, 100), (200, 210)], 3, MatchEventSubtype.extra_intron_flanking_right),
                              ([1, 1, 0], (0, 300), [(50, 60), (82, 100), (200, 292)], 3, MatchEventSubtype.fake_terminal_exon_right)])
    def test_add_extra_out_exon_events(self, read_intron_read_profile, read_region, read_introns, delta, expected):
        assigner = LongReadAssigner(self.gene_info, Params(delta), self.string_pools)
        assigner.params.max_fake_terminal_exon_len = 10
        events = []
        assigner.intron_comparator.add_extra_out_exon_events(events, read_intron_read_profile, read_region, read_introns, 100)
        assert events[0].event_type == expected

    @pytest.mark.parametrize("read_region, junctions, region, delta, expected",
                             [((0, 300), [(50, 60), (82, 100), (200, 210)], (0, 0), 3, True),
                              ((0, 300), [(48, 61), (80, 98), (200, 210)], (0, 1), 3, True),
                              ((0, 300), [(38, 61), (80, 98), (160, 220)], (0, 0), 3, False),
                              ((50, 300), [(58, 76), (80, 97), (130, 210)], (0, 1), 3, False)])
    def test_profile_for_junctions_introns(self, read_region, junctions, region, delta, expected):
        assigner = LongReadAssigner(self.gene_info, Params(delta), self.string_pools)
        assigner.params.max_suspicious_intron_abs_len = 20
        assigner.params.max_suspicious_intron_rel_len = 0.2
        assert assigner.intron_comparator.are_suspicious_introns(read_region, junctions, region) == expected

    @pytest.mark.parametrize("read_junctions, read_region, isoform_junctions, isoform_region, delta",
                             [([], (20, 200), [], (20, 290), 1)])
    def test_monoexon_match(self, read_junctions, read_region, isoform_junctions, isoform_region, delta):
        assigner = LongReadAssigner(self.gene_info, Params(delta), self.string_pools)
        match_events = assigner.intron_comparator.compare_junctions(read_junctions, read_region,
                                                                    isoform_junctions, isoform_region)
        assert len(match_events) == 1
        assert match_events[0].event_type == MatchEventSubtype.mono_exon_match

    @pytest.mark.parametrize("read_junctions, read_region, isoform_junctions, isoform_region, delta",
                             [([], (20, 200), [(150, 170)], (20, 290), 1)])
    def test_unspliced_intron_retention(self, read_junctions, read_region, isoform_junctions, isoform_region, delta):
        assigner = LongReadAssigner(self.gene_info, Params(delta), self.string_pools)
        match_events = assigner.intron_comparator.compare_junctions(read_junctions, read_region,
                                                                    isoform_junctions, isoform_region)
        assert len(match_events) == 1
        assert match_events[0].event_type == MatchEventSubtype.unspliced_intron_retention

    @pytest.mark.parametrize("read_junctions, read_region, isoform_junctions, isoform_region, delta",
                             [([], (20, 100), [(50, 170)], (20, 290), 1)])
    def test_incomplete_intron_retention_right(self, read_junctions, read_region, isoform_junctions, isoform_region, delta):
        assigner = LongReadAssigner(self.gene_info, Params(delta), self.string_pools)
        match_events = assigner.intron_comparator.compare_junctions(read_junctions, read_region,
                                                                    isoform_junctions, isoform_region)
        assert len(match_events) == 1
        assert match_events[0].event_type == MatchEventSubtype.incomplete_intron_retention_right

    @pytest.mark.parametrize("read_junctions, read_region, isoform_junctions, isoform_region, delta",
                             [([], (150, 320), [(50, 170)], (20, 290), 1)])
    def test_incomplete_intron_retention_left(self, read_junctions, read_region, isoform_junctions, isoform_region, delta):
        assigner = LongReadAssigner(self.gene_info, Params(delta), self.string_pools)
        match_events = assigner.intron_comparator.compare_junctions(read_junctions, read_region,
                                                                    isoform_junctions, isoform_region)
        assert len(match_events) == 1
        assert match_events[0].event_type == MatchEventSubtype.incomplete_intron_retention_left

    @pytest.mark.parametrize("read_junctions, read_region, isoform_junctions, isoform_region, delta",
                             [([], (100, 150), [(50, 170)], (20, 290), 1),
                              ([], (20, 55), [(50, 170)], (20, 290), 1)])
    def test_mono_exonic(self, read_junctions, read_region, isoform_junctions, isoform_region, delta):
        assigner = LongReadAssigner(self.gene_info, Params(delta), self.string_pools)
        match_events = assigner.intron_comparator.compare_junctions(read_junctions, read_region,
                                                                    isoform_junctions, isoform_region)
        assert len(match_events) == 1
        assert match_events[0].event_type == MatchEventSubtype.mono_exonic

    @pytest.mark.parametrize("read_junctions, read_region, isoform_junctions, isoform_region, delta",
                             [([(1, 10), (15,  20)], (0, 30), [(2, 10), (15,  19)], (0, 40), 3),
                              ([(1, 10), (15, 20)], (0, 30), [(1, 10), (15, 20)], (0, 50), 0),
                              ([(1, 10), (15, 20)], (0, 30), [(1, 10), (15, 21)], (0, 30), 1),
                              ([(15, 20), (25, 35)], (10, 40), [(1, 10), (15, 21), (25, 34)], (0, 40), 2)])
    def test_no_contradiction(self, read_junctions, read_region, isoform_junctions, isoform_region, delta):
        assigner = LongReadAssigner(self.gene_info, Params(delta), self.string_pools)
        match_events = assigner.intron_comparator.compare_junctions(read_junctions, read_region,
                                                                    isoform_junctions, isoform_region)
        assert len(match_events) == 1
        assert match_events[0].event_type == MatchEventSubtype.none

    @pytest.mark.parametrize("read_junctions, read_region, isoform_junctions, isoform_region, delta, expected_len",
                             [([(1, 100), (150,  200)], (0, 300), [(2, 101)], (0, 120), 1, 1),
                              ([(1, 100), (150, 200), (250, 360), (380, 390)], (0, 400), [(3, 100), (150, 201)], (0, 249), 3, 2)])
    def test_extra_intron_out_right(self, read_junctions, read_region, isoform_junctions, isoform_region, delta, expected_len):
        assigner = LongReadAssigner(self.gene_info, Params(delta), self.string_pools)
        match_events = assigner.intron_comparator.compare_junctions(read_junctions, read_region,
                                                                    isoform_junctions, isoform_region)
        assert len(match_events) == expected_len
        assert match_events[0].event_type == MatchEventSubtype.extra_intron_flanking_right

    @pytest.mark.parametrize("read_junctions, read_region, isoform_junctions, isoform_region, delta, expected_len",
                             [([(1, 100), (150, 200)], (0, 300), [(150, 201)], (110, 220), 1, 1),
                              ([(1, 100), (150, 200), (250, 360)], (0, 400), [(251, 361)], (201, 405), 3, 2)])
    def test_extra_intron_out_left(self, read_junctions, read_region, isoform_junctions, isoform_region, delta, expected_len):
        assigner = LongReadAssigner(self.gene_info, Params(delta), self.string_pools)
        match_events = assigner.intron_comparator.compare_junctions(read_junctions, read_region,
                                                                    isoform_junctions, isoform_region)
        assert len(match_events) == expected_len
        assert match_events[0].event_type == MatchEventSubtype.extra_intron_flanking_left

    @pytest.mark.parametrize("read_junctions, read_region, isoform_junctions, isoform_region, delta, expected",
                             [([(20, 50), (60, 100), (150,  200)], (0, 300), [(20, 51), (150, 201)], (0, 290), 2,
                               MatchEventSubtype.extra_intron_novel),
                              ([(20, 40), (50, 60), (150, 200)], (0, 300), [(20, 41), (150, 201)], (0, 290), 1,
                               MatchEventSubtype.extra_intron_known),
                              ([(20, 40), (78, 112), (150, 200)], (0, 300), [(20, 41), (150, 201)], (0, 290), 3,
                               MatchEventSubtype.extra_intron_known),
                              ([(20, 40), (98, 102), (150, 200)], (0, 300), [(20, 41), (150, 201)], (0, 290), 3,
                               MatchEventSubtype.none),
                              ([(20, 50), (60, 100), (150, 200)], (15, 300), [(60, 100), (150, 201)], (45, 290), 2,
                               MatchEventSubtype.fake_terminal_exon_left),
                              ([(20, 50), (60, 100), (150, 200)], (0, 205), [(20, 50), (60, 100)], (0, 170), 2,
                               MatchEventSubtype.fake_terminal_exon_right)
                              ])
    def test_extra_intron(self, read_junctions, read_region, isoform_junctions, isoform_region, delta, expected):
        assigner = LongReadAssigner(self.gene_info, Params(delta), self.string_pools)
        assigner.params.max_suspicious_intron_abs_len = 10
        assigner.params.max_suspicious_intron_rel_len = 0.1
        assigner.params.max_fake_terminal_exon_len = 10
        match_events = assigner.intron_comparator.compare_junctions(read_junctions, read_region,
                                                                    isoform_junctions, isoform_region)
        assert len(match_events) == 1
        assert match_events[0].event_type == expected

    @pytest.mark.parametrize("read_junctions, read_region, isoform_junctions, isoform_region, delta",
                             [([(10, 50), (150,  200)], (0, 300), [(10, 51), (80, 100), (150, 200), (225, 240)], (0, 310), 3)])
    def test_missed_intron_in(self, read_junctions, read_region, isoform_junctions, isoform_region, delta):
        assigner = LongReadAssigner(self.gene_info, Params(delta), self.string_pools)
        match_events = assigner.intron_comparator.compare_junctions(read_junctions, read_region,
                                                                    isoform_junctions, isoform_region)
        event_types = [match_events[i].event_type for i in range(len(match_events))]
        assert len(match_events) == 2
        assert set(event_types) == {MatchEventSubtype.intron_retention}

    @pytest.mark.parametrize("read_junctions, read_region, isoform_junctions, isoform_region, delta, expected",
                             [([(10, 50)], (0, 100), [(10, 25), (40, 49)], (0, 99), 3, MatchEventSubtype.exon_skipping_novel),
                              ([(80, 110)], (50, 150), [(80, 90), (105, 110)], (50, 149), 3, MatchEventSubtype.exon_skipping_known),
                              ([(80, 110)], (20, 150), [(65, 90), (105, 110)], (20, 149), 3, MatchEventSubtype.exon_merge_known),
                              ([(70, 110)], (20, 150), [(55, 90), (105, 110)], (20, 149), 3, MatchEventSubtype.exon_merge_novel),
                              ([(81, 109)], (50, 150), [(80, 98), (102, 110)], (50, 149), 3, MatchEventSubtype.exon_misalignment)])
    def test_exon_skipping(self, read_junctions, read_region, isoform_junctions, isoform_region, delta, expected):
        assigner = LongReadAssigner(self.gene_info, Params(delta), self.string_pools)
        match_events = assigner.intron_comparator.compare_junctions(read_junctions, read_region,
                                                                    isoform_junctions, isoform_region)
        assert len(match_events) == 1
        assert match_events[0].event_type == expected

    @pytest.mark.parametrize("read_junctions, read_region, isoform_junctions, isoform_region, delta, expected",
                             [([(10, 30), (50, 100)], (0, 120), [(10, 60), (80, 100)], (0, 119), 3,
                               MatchEventSubtype.mutually_exclusive_exons_novel),
                              ([(50, 60), (81, 149)], (0, 200), [(50, 100), (119, 150)], (0, 219), 3,
                               MatchEventSubtype.mutually_exclusive_exons_known),
                              ([(50, 60), (81, 130), (140, 181)], (0, 200), [(50, 100), (119, 200)], (0, 219), 3,
                               MatchEventSubtype.alternative_structure_novel),
                              ([(50, 60), (81, 150), (200, 210)], (0, 250), [(50, 99), (115, 210)], (0, 279), 3,
                               MatchEventSubtype.alternative_structure_known)
                              ])
    def test_complex_alteration(self, read_junctions, read_region, isoform_junctions, isoform_region, delta, expected):
        assigner = LongReadAssigner(self.gene_info, Params(delta), self.string_pools)
        match_events = assigner.intron_comparator.compare_junctions(read_junctions, read_region,
                                                                    isoform_junctions, isoform_region)
        assert len(match_events) == 1
        assert match_events[0].event_type == expected

    @pytest.mark.parametrize("read_junctions, read_region, isoform_junctions, isoform_region, delta, expected",
                             [([(10, 30), (55, 100)], (0, 150), [(10, 100)], (9, 149), 2,
                               MatchEventSubtype.exon_gain_novel),
                              ([(50, 60), (80, 100)], (0, 150), [(50, 100)], (9, 149), 2,
                               MatchEventSubtype.exon_gain_known),
                              ([(10, 30), (55, 100)], (0, 150), [(10, 75)], (9, 149), 2,
                               MatchEventSubtype.exon_detach_novel),
                              ([(50, 60), (80, 110)], (0, 150), [(50, 90)], (9, 149), 2,
                               MatchEventSubtype.exon_detach_known),
                              ([(50, 60), (80, 110)], (20, 150), [(20, 60), (80, 110)], (0, 150), 2,
                               MatchEventSubtype.terminal_exon_shift_known),
                              ([(45, 60), (80, 110)], (22, 150), [(10, 60), (80, 110)], (0, 150), 2,
                               MatchEventSubtype.terminal_exon_shift_novel),
                              ([(50, 60), (80, 110)], (30, 150), [(20, 60), (80, 110)], (0, 150), 2,
                               MatchEventSubtype.terminal_exon_misalignment_left),
                              ([(10, 13), (95, 100)], (0, 150), [(10, 100)], (9, 149), 2,
                               MatchEventSubtype.intron_retention)])
    def test_exon_gain(self, read_junctions, read_region, isoform_junctions, isoform_region, delta, expected):
        assigner = LongReadAssigner(self.gene_info, Params(delta), self.string_pools)
        assigner.params.max_suspicious_intron_abs_len = 10
        assigner.params.max_suspicious_intron_rel_len = 0.1
        match_events = assigner.intron_comparator.compare_junctions(read_junctions, read_region,
                                                                    isoform_junctions, isoform_region)
        assert len(match_events) == 1
        assert match_events[0].event_type == expected

    @pytest.mark.parametrize("read_junctions, read_region, isoform_junctions, isoform_region, delta, expected",
                             [([(50, 70), (80, 100)], (0, 150), [(50, 60), (80, 100)], (9, 149), 2,
                               MatchEventSubtype.alt_right_site_novel),
                              ([(50, 60), (80, 110)], (0, 150), [(50, 60), (80, 100)], (9, 149), 2,
                               MatchEventSubtype.alt_right_site_known),
                              ([(50, 60), (90, 100)], (0, 150), [(50, 60), (80, 100)], (9, 149), 2,
                               MatchEventSubtype.alt_left_site_novel),
                              ([(50, 60), (80, 100)], (0, 150), [(40, 60), (80, 100)], (9, 149), 2,
                               MatchEventSubtype.alt_left_site_known),
                              ([(50, 60), (80, 100)], (0, 150), [(50, 60), (95, 115)], (9, 149), 2,
                               MatchEventSubtype.intron_migration),
                              ([(50, 60), (85, 105)], (0, 150), [(50, 60), (80, 100)], (9, 149), 2,
                               MatchEventSubtype.intron_shift),
                              ([(50, 60), (68, 85)], (0, 150), [(50, 60), (80, 100)], (9, 149), 2,
                               MatchEventSubtype.intron_alternation_novel),
                              ([(50, 60), (75, 115)], (0, 150), [(50, 60), (80, 100)], (9, 149), 2,
                               MatchEventSubtype.intron_alternation_novel),
                              ([(50, 60), (80, 110)], (0, 150), [(50, 60), (65, 100)], (9, 149), 2,
                               MatchEventSubtype.intron_alternation_known),
                              ([(50, 60), (98, 100)], (0, 150), [(50, 60), (80, 100)], (9, 149), 2,
                               MatchEventSubtype.intron_retention),
                              ([(50, 60), (93, 96)], (0, 150), [(50, 60), (80, 100)], (9, 149), 2,
                               MatchEventSubtype.intron_retention)
                              ])
    def test_single_introns(self, read_junctions, read_region, isoform_junctions, isoform_region, delta, expected):
        assigner = LongReadAssigner(self.gene_info, Params(delta), self.string_pools)
        assigner.params.max_suspicious_intron_abs_len = 5
        assigner.params.max_suspicious_intron_rel_len = 0.1
        match_events = assigner.intron_comparator.compare_junctions(read_junctions, read_region,
                                                                    isoform_junctions, isoform_region)
        assert len(match_events) == 1
        assert match_events[0].event_type == expected

    @pytest.mark.parametrize("read_junctions, read_region, isoform_junctions, isoform_region, delta, expected",
                             [([(80, 100)], (0, 150), [(50, 60), (80, 100)], (9, 149), 2,
                               MatchEventSubtype.intron_retention),
                              ([(50, 60)], (0, 95), [(50, 60), (80, 100)], (9, 149), 2,
                               MatchEventSubtype.incomplete_intron_retention_right),
                              ([(80, 100)], (40, 150), [(20, 60), (80, 100)], (9, 149), 2,
                               MatchEventSubtype.incomplete_intron_retention_left),
                              ])
    def test_ir(self, read_junctions, read_region, isoform_junctions, isoform_region, delta, expected):
        assigner = LongReadAssigner(self.gene_info, Params(delta), self.string_pools)
        assigner.params.max_suspicious_intron_abs_len = 5
        assigner.params.max_suspicious_intron_rel_len = 0.1
        match_events = assigner.intron_comparator.compare_junctions(read_junctions, read_region,
                                                                    isoform_junctions, isoform_region)
        assert len(match_events) == 1
        assert match_events[0].event_type == expected

    @pytest.mark.parametrize("read_junctions, read_region, isoform_junctions, isoform_region, delta",
                             [([(50, 60)], (0, 85), [(50, 60), (80, 100)], (9, 149), 2),
                              ([(80, 100)], (57, 150), [(20, 60), (80, 100)], (9, 149), 2),
                              ])
    def test_no_ir(self, read_junctions, read_region, isoform_junctions, isoform_region, delta):
        assigner = LongReadAssigner(self.gene_info, Params(delta), self.string_pools)
        assigner.params.max_suspicious_intron_abs_len = 5
        assigner.params.max_suspicious_intron_rel_len = 0.1
        match_events = assigner.intron_comparator.compare_junctions(read_junctions, read_region,
                                                                    isoform_junctions, isoform_region)
        assert match_events[0].event_type == MatchEventSubtype.none


class TestDetectContradictionType:
    gene_info = GeneInfoTuple(IntronProfiles([(1, 1)]), 1, 100)
    string_pools = StringPoolManager()

    def test(self):
        assigner = LongReadAssigner(self.gene_info, Params(3), self.string_pools)
        match_events = assigner.intron_comparator.detect_contradiction_type((0, 200), [(50, 75)], (0, 200), [(45, 70)], [((0, 0), (0, 0))])
        assert len(match_events) == 1
        assert match_events[0].event_type == MatchEventSubtype.intron_shift


class TestAssignIsoform:
    params = Params(3)
    source_dir = os.path.dirname(os.path.realpath(__file__))
    gffutils_db = gffutils.FeatureDB(os.path.join(source_dir, 'toy_data/synth.db'), keep_order=True)
    gene_db = gffutils_db['ENSMUSG00000020196.10']
    gene_info = GeneInfo([gene_db], gffutils_db)
    string_pools = StringPoolManager()
    assigner = LongReadAssigner(gene_info, params, string_pools)
    gene_region = (gene_info.start, gene_info.end)

    intron_profile_constructor = \
        OverlappingFeaturesProfileConstructor(gene_info.intron_profiles.features, gene_region,
                                              comparator=partial(equal_ranges, delta=params.delta),
                                              absence_condition=partial(overlaps_at_least,
                                                                        delta=params.minimal_intron_absence_overlap),
                                              delta=params.delta)
    exon_profile_constructor = \
        OverlappingFeaturesProfileConstructor(gene_info.exon_profiles.features, gene_region,
                                              comparator=partial(equal_ranges, delta=params.delta),
                                              delta=params.delta)
    split_exon_profile_constructor = \
        NonOverlappingFeaturesProfileConstructor(gene_info.split_exon_profiles.features,
                                                 comparator=partial(overlaps_at_least,
                                                                    delta=params.minimal_exon_overlap),
                                                 delta=params.delta)

    @pytest.mark.parametrize("sorted_blocks, polya_pos, polyt_pos, isoform_id, expected_event",
                             [([(1000, 1100), (2000, 2100), (2300, 2400), (3000, 3300), (9500, 10000)], -1, -1,
                               "ENSMUST00000001712.7", MatchEventSubtype.fsm),
                              ([(1000, 1100), (2000, 2100), (2300, 2400), (3000, 3300)], -1, -1,
                               "ENSMUST00000001712.7", MatchEventSubtype.ism_right),
                              ([(2000, 2098), (2301, 2400), (3001, 3300), (9500, 10003)], -1, -1,
                               "ENSMUST00000001712.7", MatchEventSubtype.ism_left),
                              ([(2000, 2100), (2300, 2400), (3000, 3300)], -1, -1,
                               "ENSMUST00000001712.7", MatchEventSubtype.ism_internal),
                              ([(1000, 1100), (2000, 2200), (2500, 2600), (3000, 3300), (6000, 6010), (9500, 10000)],
                               -1, -1, "ENSMUST00000001713.7", MatchEventSubtype.fsm),
                              ([(1040, 1100), (2000, 2201), (2500, 2602), (2998, 3300), (6000, 6011), (9500, 9600)],
                               -1, -1, "ENSMUST00000001713.7", MatchEventSubtype.fsm),
                              ([(7998, 8201), (8500, 8800)], -1, -1,
                               "ENSMUST00000001715.7", MatchEventSubtype.fsm),
                              ([(7000, 7500)], -1, -1,
                               "ENSMUST00000001714.7", MatchEventSubtype.mono_exon_match),
                              ([(7100, 7300)], -1, -1,
                               "ENSMUST00000001714.7", MatchEventSubtype.mono_exon_match),
                              ([(6999, 7503)], -1, -1,
                               "ENSMUST00000001714.7", MatchEventSubtype.mono_exon_match),
                              ])
    def test_assign_unique(self, sorted_blocks, polya_pos, polyt_pos, isoform_id, expected_event):
        intron_profile = self.intron_profile_constructor.construct_intron_profile(sorted_blocks, polya_pos, polyt_pos)
        exon_profile = self.exon_profile_constructor.construct_exon_profile(sorted_blocks, polya_pos, polyt_pos)
        split_exon_profile = self.split_exon_profile_constructor.construct_profile(sorted_blocks, polya_pos, polyt_pos)
        combined_profile = CombinedReadProfiles(intron_profile, exon_profile, split_exon_profile,
                                                PolyAInfo(polya_pos, polyt_pos, -1, -1))

        read_assignment = self.assigner.assign_to_isoform("read_id", combined_profile)
        assert read_assignment.assignment_type == ReadAssignmentType.unique
        assert len(read_assignment.isoform_matches) == 1
        assert read_assignment.isoform_matches[0].assigned_transcript == isoform_id
        assert expected_event in {e.event_type for e in read_assignment.isoform_matches[0].match_subclassifications}

    @pytest.mark.parametrize("sorted_blocks, polya_pos, polyt_pos, isoform_id, expected_event",
                             [([(1000, 1108), (2007, 2100), (2300, 2400), (3000, 3300), (9500, 10000)], -1, -1,
                               "ENSMUST00000001712.7", MatchEventSubtype.intron_shift),
                              ([(1000, 1100), (2000, 2092), (2295, 2400), (3000, 3300)], -1, -1,
                               "ENSMUST00000001712.7", MatchEventSubtype.intron_shift),
                              ([(2000, 2098), (2301, 2400), (3001, 3300), (9500, 10012)], -1, -1,
                               "ENSMUST00000001712.7", MatchEventSubtype.exon_elongation_right),
                              ([(1000, 1100), (2000, 2200), (2500, 2600), (3000, 3303), (9496, 10000)],
                               -1, -1, "ENSMUST00000001713.7", MatchEventSubtype.exon_misalignment),
                              ([(988, 1100), (2000, 2200), (2500, 2602), (2998, 3300), (6000, 6011), (9500, 9600)],
                               -1, -1, "ENSMUST00000001713.7", MatchEventSubtype.exon_elongation_left),
                              ([(7998, 8201), (8500, 8800), (9900, 9907)], -1, -1,
                               "ENSMUST00000001715.7", MatchEventSubtype.fake_terminal_exon_right),
                              ([(4000, 4010), (7998, 8201), (8500, 8800)], -1, -1,
                               "ENSMUST00000001715.7", MatchEventSubtype.fake_terminal_exon_left),
                              ])
    def test_assign_unique_minor_error(self, sorted_blocks, polya_pos, polyt_pos, isoform_id, expected_event):
        self.assigner.params.max_missed_exon_len = 20
        self.assigner.params.max_fake_terminal_exon_len = 20
        intron_profile = self.intron_profile_constructor.construct_intron_profile(sorted_blocks, polya_pos, polyt_pos)
        exon_profile = self.exon_profile_constructor.construct_exon_profile(sorted_blocks, polya_pos, polyt_pos)
        split_exon_profile = self.split_exon_profile_constructor.construct_profile(sorted_blocks, polya_pos, polyt_pos)
        combined_profile = CombinedReadProfiles(intron_profile, exon_profile, split_exon_profile,
                                                PolyAInfo(polya_pos, polyt_pos, -1, -1))

        read_assignment = self.assigner.assign_to_isoform("read_id", combined_profile)
        assert read_assignment.assignment_type == ReadAssignmentType.unique_minor_difference
        assert len(read_assignment.isoform_matches) == 1
        assert read_assignment.isoform_matches[0].assigned_transcript == isoform_id
        assert expected_event in {e.event_type for e in read_assignment.isoform_matches[0].match_subclassifications}

    @pytest.mark.parametrize("sorted_blocks, polya_pos, polyt_pos, isoform_id",
                             [([(1000, 1108), (2007, 2100)], -1, -1,
                               "ENSMUST00000001712.7"),
                              ([(1050, 1100), (2001, 2092)], -1, -1,
                               "ENSMUST00000001712.7"),
                              ])
    def test_assign_unique_ambiguous(self, sorted_blocks, polya_pos, polyt_pos, isoform_id):
        self.assigner.params.max_missed_exon_len = 20
        self.assigner.params.max_fake_terminal_exon_len = 20
        intron_profile = self.intron_profile_constructor.construct_intron_profile(sorted_blocks, polya_pos, polyt_pos)
        exon_profile = self.exon_profile_constructor.construct_exon_profile(sorted_blocks, polya_pos, polyt_pos)
        split_exon_profile = self.split_exon_profile_constructor.construct_profile(sorted_blocks, polya_pos, polyt_pos)
        combined_profile = CombinedReadProfiles(intron_profile, exon_profile, split_exon_profile,
                                                PolyAInfo(polya_pos, polyt_pos, -1, -1))

        read_assignment = self.assigner.assign_to_isoform("read_id", combined_profile)
        assert read_assignment.assignment_type == ReadAssignmentType.ambiguous
        assert len(read_assignment.isoform_matches) > 1
        assert isoform_id in {im.assigned_transcript for im in read_assignment.isoform_matches}

    @pytest.mark.parametrize("sorted_blocks, polya_pos, polyt_pos, isoform_id, expected_event",
                             [([(1000, 1200), (2000, 2100), (2300, 2400), (3000, 3300), (9500, 10000)], -1, -1,
                               "ENSMUST00000001712.7", MatchEventSubtype.alt_left_site_novel),
                              ([(1000, 1100), (2000, 2400), (3000, 3300), (9500, 10000)], -1, -1,
                               "ENSMUST00000001712.7", MatchEventSubtype.intron_retention),
                              ([(500, 600), (1000, 1100), (2000, 2100), (2300, 2400), (3000, 3300), (9500, 10000)], -1, -1,
                               "ENSMUST00000001712.7", MatchEventSubtype.extra_intron_flanking_left),
                              ([(1000, 1100), (2000, 2100), (2300, 2400), (2700, 2750), (3000, 3300)], -1, -1,
                               "ENSMUST00000001712.7", MatchEventSubtype.exon_gain_novel),
                              ([(2000, 2098), (2301, 2400), (3001, 3300), (9500, 11003)], -1, -1,
                               "ENSMUST00000001712.7", MatchEventSubtype.major_exon_elongation_right),
                              ([(1000, 1100), (2000, 2200), (2500, 2600), (3000, 3300), (5000, 5010), (9500, 10000)],
                               -1, -1, "ENSMUST00000001713.7", MatchEventSubtype.mutually_exclusive_exons_novel),
                              ([(1040, 1100), (2000, 2201), (2500, 2602), (2998, 3300), (6000, 6011), (9500, 9600), (9700, 9999)],
                               -1, -1, "ENSMUST00000001713.7", MatchEventSubtype.extra_intron_novel),
                              ([(7998, 8201), (8300, 8333), (8500, 8800)], -1, -1,
                               "ENSMUST00000001715.7", MatchEventSubtype.exon_gain_novel),
                              ([(7100, 7800)], -1, -1,
                               "ENSMUST00000001714.7", MatchEventSubtype.major_exon_elongation_right),
                              ([(1040, 1100), (2000, 2201), (2500, 2602), (2998, 3900)], -1, -1,
                               "ENSMUST00000001713.7", MatchEventSubtype.incomplete_intron_retention_right),
                              ([(2500, 6010)], -1, -1,
                               "ENSMUST00000001713.7", MatchEventSubtype.unspliced_intron_retention),
                              ([(900, 1100), (2000, 2100), (2300, 2400), (3000, 3300), (9500, 10000)], -1, 900,
                               "ENSMUST00000001712.7", MatchEventSubtype.alternative_polya_site_left),
                              ([(1000, 1100), (1900, 2100)], -1, -1,
                               "ENSMUST00000001713.7", MatchEventSubtype.alt_right_site_novel),
                              ([(1000, 2100)], -1, -1,
                               "ENSMUST00000001713.7", MatchEventSubtype.unspliced_intron_retention),
                              ([(1000, 1100), (2000, 3300)], -1, -1,
                               "ENSMUST00000001713.7", MatchEventSubtype.intron_retention),
                              ])
    def test_assign_inconsistent(self, sorted_blocks, polya_pos, polyt_pos, isoform_id, expected_event):
        self.assigner.params.max_missed_exon_len = 20
        self.assigner.params.max_fake_terminal_exon_len = 20
        intron_profile = self.intron_profile_constructor.construct_intron_profile(sorted_blocks, polya_pos, polyt_pos)
        exon_profile = self.exon_profile_constructor.construct_exon_profile(sorted_blocks, polya_pos, polyt_pos)
        split_exon_profile = self.split_exon_profile_constructor.construct_profile(sorted_blocks, polya_pos, polyt_pos)
        combined_profile = CombinedReadProfiles(intron_profile, exon_profile, split_exon_profile,
                                                PolyAInfo(polya_pos, polyt_pos, -1, -1))

        read_assignment = self.assigner.assign_to_isoform("read_id", combined_profile)
        assert read_assignment.assignment_type.is_inconsistent()
        if isoform_id:
            assert isoform_id in {im.assigned_transcript for im in read_assignment.isoform_matches}
        if expected_event:
            assert expected_event in {e.event_type for e in read_assignment.isoform_matches[0].match_subclassifications}



############################################################################
# Copyright (c) 2025-2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################


import os


def convert_chr_id_to_file_name_str(chr_id: str):
    return chr_id.replace('/', '_')


def reads_collected_lock_file_name(sample_out_raw, chr_id):
    return "{}_{}_collected".format(sample_out_raw, convert_chr_id_to_file_name_str(chr_id))


def reads_processed_lock_file_name(dump_filename, chr_id):
    chr_dump_file = dump_filename + "_" + convert_chr_id_to_file_name_str(chr_id)
    return "{}_processed".format(chr_dump_file)


def read_group_lock_filename(sample):
    return sample.read_group_file + "_lock"


def split_barcodes_lock_filename(sample):
    return sample.barcodes_split_reads + "_lock"


def clean_locks(chr_ids, base_name, fname_function):
    for chr_id in chr_ids:
        fname = fname_function(base_name, chr_id)
        if os.path.exists(fname):
            os.remove(fname)


def saves_file_name(out_raw_file: str, chr_id: str):
    return out_raw_file + "_" + convert_chr_id_to_file_name_str(chr_id)


def multimappers_file_name(out_raw_file: str, chr_id: str):
    return out_raw_file + "_multimappers_" + convert_chr_id_to_file_name_str(chr_id)


def filtered_reads_file_name(out_raw_file: str, chr_id: str):
    return out_raw_file + "_filtered_" + chr_id


def umi_filtered_reads_file_name(out_umi_filtered_tmp: str, chr_id:str, edit_distance:int):
    return out_umi_filtered_tmp + ("_%s_ED%d" % (chr_id, edit_distance))


def umi_filtered_lock_file_name(out_umi_filtered_done: str, chr_id: str, edit_distance: int):
    return out_umi_filtered_done + ("_%s_ED%d" % (chr_id, edit_distance))


def allinfo_file_name(out_umi_filtered_tmp: str, chr_id: str, edit_distance: int):
    return umi_filtered_reads_file_name(out_umi_filtered_tmp, chr_id, edit_distance) + ".allinfo"


def allinfo_stats_file_name(out_umi_filtered_tmp: str, chr_id: str, edit_distance: int):
    return umi_filtered_reads_file_name(out_umi_filtered_tmp, chr_id, edit_distance) + ".stats.tsv"


def umi_filtered_global_lock_file_name(out_umi_filtered_done: str):
    return out_umi_filtered_done + ".lock"


def umi_barcode2barcode_prefix(out_umi_filtered_done: str, col_index: int):
    """Get prefix for barcode2barcode UMI filtering output (per spot column)."""
    return out_umi_filtered_done + ".barcode_barcode_col%d" % col_index


def umi_barcode2barcode_global_lock(out_umi_filtered_done: str, col_index: int):
    """Lock file for barcode2barcode UMI filtering of a specific column."""
    return umi_barcode2barcode_prefix(out_umi_filtered_done, col_index) + ".lock"


def dynamic_pools_file_name(out_raw_file: str, chr_id: str):
    """File name for per-chromosome dynamic string pools (for read groups from BAM tags/read IDs)."""
    return out_raw_file + "_dynamic_pools_" + convert_chr_id_to_file_name_str(chr_id)


def barcode_pools_file_name(out_raw_file: str, chr_id: str):
    """File name for per-chromosome barcode/UMI pools (for --barcoded_bam mode)."""
    return out_raw_file + "_barcode_pools_" + convert_chr_id_to_file_name_str(chr_id)


# Read collection auxiliary files

def read_groups_file_name(save_file: str):
    """File name for storing read group assignments during collection."""
    return save_file + "_groups"


def bamstat_file_name(save_file: str):
    """File name for BAM alignment statistics."""
    return save_file + "_bamstat"


def info_file_name(out_raw_file: str):
    """File name for sample collection info (resume functionality)."""
    return out_raw_file + "_info"


def collection_lock_file_name(out_raw_file: str):
    """File name for sample collection lock (resume functionality)."""
    return out_raw_file + "_lock"


# Model construction auxiliary files

def read_stat_file_name(saves_file: str):
    """File name for read assignment statistics."""
    return saves_file + "_read_stat"


def transcript_stat_file_name(saves_file: str):
    """File name for transcript model statistics."""
    return saves_file + "_transcript_stat"


def read_group_file_from_saves(saves_prefix: str):
    """Convert saves file prefix to read_group file prefix."""
    if saves_prefix.endswith('.save'):
        return saves_prefix.replace('.save', '.read_group')
    return None


# UMI filtering output files

def umi_output_prefix(out_umi_filtered: str, edit_distance: int):
    """Get UMI filtered output prefix with edit distance encoding."""
    if edit_distance < 0:
        return out_umi_filtered + ".ALL"
    return out_umi_filtered + ".ED%d" % edit_distance


# Counter output files

def counts_prefix(output_prefix: str):
    """Get counts file prefix from output prefix."""
    return output_prefix + "_counts"


def tpm_prefix(output_prefix: str):
    """Get TPM file prefix from output prefix."""
    return output_prefix + "_tpm"


def counts_file_name(counts_prefix: str, linear: bool = False):
    """Get counts file name with optional linear format."""
    return counts_prefix + (".linear.tsv" if linear else ".tsv")


def tpm_file_name(tpm_prefix: str):
    """Get TPM file name."""
    return tpm_prefix + ".tsv"


def counts_stats_file_name(counts_file: str):
    """Get stats file name from counts file."""
    return counts_file + ".stats"


def counts_usable_file_name(counts_file: str):
    """Get usable reads file name from counts file."""
    return counts_file + ".usable"


# Matrix Market format files (for grouped counts)

def mtx_matrix_file(output_prefix: str):
    """Get Matrix Market file name."""
    return output_prefix + ".matrix.mtx"


def mtx_features_file(output_prefix: str):
    """Get features file name for Matrix Market format."""
    return output_prefix + ".features.tsv"


def mtx_barcodes_file(output_prefix: str):
    """Get barcodes file name for Matrix Market format."""
    return output_prefix + ".barcodes.tsv"